/*----------------------------------------------------------------------------

	drawutil.c

	This reusable module contains miscellaneous drawing utility routines.
	
	Copyright  1994, Northwestern University.

----------------------------------------------------------------------------*/

#include <GestaltEqu.h>
#include <Palettes.h>

#include "def.h"
#include "drawutil.h"



static DeviceLoopDrawingUPP gDrawGrayRectOneDepthUPP = nil;
static DeviceLoopDrawingUPP gDrawGrayRoundRectOneDepthUPP = nil;
static DeviceLoopDrawingUPP gDrawColorPictOneDepthUPP = nil;
static DeviceLoopDrawingUPP gFillColorPolyOneDepthUPP = nil;



/*----------------------------------------------------------------------------
	HasColorQD 
	
	Check to see if color QuickDraw is available.
	
	Exit:	function result = true if we have color QD.
----------------------------------------------------------------------------*/

Boolean HasColorQD (void)
{
	OSErr err = noErr;
	long quickdrawVersion;
	static Boolean haveChecked = false;
	static Boolean haveColorQD;
	
	if (!haveChecked) {
		err = Gestalt(gestaltQuickdrawVersion, &quickdrawVersion);
		haveColorQD = err == noErr && quickdrawVersion >= gestalt8BitQD;
	}
	return haveColorQD;
}



/*----------------------------------------------------------------------------
	SetPortTextStyle 
	
	Set the font, size, and style of the current port.
			
	Entry:	*style = text style record.
----------------------------------------------------------------------------*/

void SetPortTextStyle (TextStyle *style)
{
	TextFont(style->tsFont);
	TextFace(style->tsFace);
	TextSize(style->tsSize);
}



/*----------------------------------------------------------------------------
	GetPortTextStyle 
	
	Get the font, size, and style of the current port.
			
	Exit:	*style = text style record.
----------------------------------------------------------------------------*/

void GetPortTextStyle (TextStyle *style)
{
	style->tsFont = qd.thePort->txFont;
	style->tsFace = qd.thePort->txFace;
	style->tsSize = qd.thePort->txSize;
}



/*----------------------------------------------------------------------------
	DrawGrayRect 
	
	Draw a gray rectangle.
			
	Entry:	r = pointer to rectangle.
----------------------------------------------------------------------------*/

static pascal void DrawGrayRectOneDepth (short depth, short deviceFlags,
	GDHandle targetDevice, long userData)
{
	RGBColor fg, bg, g;
	
	if (targetDevice != nil) {
		GetForeColor(&fg);
		GetBackColor(&bg);
		g = fg;
		if (GetGray(targetDevice, &bg, &g)) {
			RGBForeColor(&g);
			FrameRect((Rect*)userData);
			RGBForeColor(&fg);
			return;
		}
	}
	PenPat(&qd.gray);
	FrameRect((Rect*)userData);
	PenPat(&qd.black);
}

void DrawGrayRect (Rect *r)
{
	DeviceLoop(qd.thePort->visRgn, gDrawGrayRectOneDepthUPP, (long)r, 0);
}



/*----------------------------------------------------------------------------
	DrawGrayRoundRect 
	
	Draw a gray round rectangle.
			
	Entry:	r = pointer to rectangle.
			ovalWidth = oval width.
			ovalHeight = oval height.
----------------------------------------------------------------------------*/

typedef struct TDrawGrayRoundRectUserData {
	Rect *r;
	short ovalWidth;
	short ovalHeight;
} TDrawGrayRoundRectUserData;

static pascal void DrawGrayRoundRectOneDepth (short depth, short deviceFlags,
	GDHandle targetDevice, long userData)
{
	RGBColor fg, bg, g;
	TDrawGrayRoundRectUserData *ud;
	
	ud = (TDrawGrayRoundRectUserData*)userData;
	if (targetDevice != nil) {
		GetForeColor(&fg);
		GetBackColor(&bg);
		g = fg;
		if (GetGray(targetDevice, &bg, &g)) {
			RGBForeColor(&g);
			FrameRoundRect(ud->r, ud->ovalWidth, ud->ovalHeight);
			RGBForeColor(&fg);
			return;
		}
	}
	PenPat(&qd.gray);
	FrameRoundRect(ud->r, ud->ovalWidth, ud->ovalHeight);
	PenPat(&qd.black);
}

void DrawGrayRoundRect (Rect *r, short ovalWidth, short ovalHeight)
{
	TDrawGrayRoundRectUserData ud;

	ud.r = r;
	ud.ovalWidth = ovalWidth;
	ud.ovalHeight = ovalHeight;
	DeviceLoop(qd.thePort->visRgn, gDrawGrayRoundRectOneDepthUPP, (long)&ud, 0);
}



/*----------------------------------------------------------------------------
	DrawColorPict 
	
	Draw a color picture.
			
	Entry:	colorID = resource id of color version of PICT.
			bwID = resource id of black and white version of PICT.
			r = pointer to rectangle.
----------------------------------------------------------------------------*/

typedef struct TDrawColorPictUserData {
	short colorID;
	short bwID;
	Rect *r;
} TDrawColorPictUserData;

static pascal void DrawColorPictOneDepth (short depth, short deviceFlags,
	GDHandle targetDevice, long userData)
{
	TDrawColorPictUserData *ud;
	
	ud = (TDrawColorPictUserData*)userData;
	if (targetDevice != nil && depth > 2) {
		DrawPicture(GetPicture(ud->colorID), ud->r);
	} else {
		DrawPicture(GetPicture(ud->bwID), ud->r);
	}
}

void DrawColorPict (short colorID, short bwID, Rect *r)
{
	TDrawColorPictUserData ud;

	ud.colorID = colorID;
	ud.bwID = bwID;
	ud.r = r;
	DeviceLoop(qd.thePort->visRgn, gDrawColorPictOneDepthUPP, (long)&ud, 0);
}



/*----------------------------------------------------------------------------
	FillColorPoly 
	
	Fill a polygon in color.
			
	Entry:	poly = handle to polygon.
			color = pointer to color.
			pat = pattern.
----------------------------------------------------------------------------*/

typedef struct TFillColorPolyUserData {
	PolyHandle poly;
	RGBColor *color;
	ConstPatternParam pat;
} TFillColorPolyUserData;

static pascal void FillColorPolyOneDepth (short depth, short deviceFlags,
	GDHandle targetDevice, long userData)
{
	TFillColorPolyUserData *ud;
	RGBColor fgColor;
	
	ud = (TFillColorPolyUserData*)userData;
	if (targetDevice != nil && depth > 2) {
		GetForeColor(&fgColor);
		RGBForeColor(ud->color);
		FillPoly(ud->poly, ud->pat);
		RGBForeColor(&fgColor);
	}
}

void FillColorPoly (PolyHandle poly, RGBColor *color, ConstPatternParam pat)
{
	TFillColorPolyUserData ud;

	ud.poly = poly;
	ud.color = color;
	ud.pat = pat;
	DeviceLoop(qd.thePort->visRgn, gFillColorPolyOneDepthUPP, (long)&ud, 0);
}



/*----------------------------------------------------------------------------
	ComplementRgn 
	
	Complement a QuickDraw region.
			
	Entry:	rgn = handle to region.
	
	Exit:	Region complemented.
----------------------------------------------------------------------------*/

void ComplementRgn (RgnHandle rgn)
{
	static RgnHandle bigRgn = nil;
	
	if (bigRgn == nil) {
		bigRgn = NewRgn();
		SetRectRgn(bigRgn, -0x7fff, -0x7fff, 0x7fff, 0x7fff);
	}
	DiffRgn(bigRgn, rgn, rgn);
}



/*----------------------------------------------------------------------------
	LocalToGlobalRect 
	
	Convert a rectangle from local to global coordinates.
			
	Entry:	r = pointer to rectangle.
----------------------------------------------------------------------------*/

void LocalToGlobalRect (Rect *r)
{
	LocalToGlobal((Point*)&r->top);
	LocalToGlobal((Point*)&r->bottom);
}



/*----------------------------------------------------------------------------
	GlobalToLocalRect 
	
	Convert a rectangle from global to local coordinates.
			
	Entry:	r = pointer to rectangle.
----------------------------------------------------------------------------*/

void GlobalToLocalRect (Rect *r)
{
	GlobalToLocal((Point*)&r->top);
	GlobalToLocal((Point*)&r->bottom);
}



/*----------------------------------------------------------------------------
	LocalToGlobalRgn 
	
	Convert a region from local to global coordinates.
			
	Entry:	rgn = handle to region.
----------------------------------------------------------------------------*/

void LocalToGlobalRgn (RgnHandle rgn)
{
	Point where;
	
	SetPt(&where, 0, 0);
	LocalToGlobal(&where);
	OffsetRgn(rgn, where.h, where.v);
}



/*----------------------------------------------------------------------------
	GlobalToLocalRgn 
	
	Convert a region from global to local coordinates.
			
	Entry:	rgn = handle to region.
----------------------------------------------------------------------------*/

void GlobalToLocalRgn (RgnHandle rgn)
{
	Point where;
	
	SetPt(&where, 0, 0);
	LocalToGlobal(&where);
	OffsetRgn(rgn, -where.h, -where.v);
}



/*----------------------------------------------------------------------------
	GetFontNumber 
	
	Get the font number corresponding to a font name (adapted from Think Ref).
			
	Entry:	fontName = font name.
	
	Exit:	function result = true if font exists.
			*fontNum = font number.
			
	If the font doesn't exist, the font number is set to Monaco, or to 
	applFont if Monaco also doesn't exist.
----------------------------------------------------------------------------*/

Boolean GetFontNumber (Str255 fontName, short *fontNum)
{
	Str255 systemFontName;

	GetFNum(fontName, fontNum);
	if (*fontNum != 0) return true;
	GetFontName(0, systemFontName);
	if (EqualString(fontName, systemFontName, false, false)) return true;
	GetFNum("\pMonaco", fontNum);
	if (*fontNum == 0) *fontNum = applFont;
	return false;
}



/*----------------------------------------------------------------------------
	GetFontLineHeight 
	
	Get the line height in pixels of a window font.
			
	Entry:	wind = pointer to window.
	
	Exit:	function result = line height.
----------------------------------------------------------------------------*/

short GetFontLineHeight (WindowPtr wind)
{
	FontInfo fInfo;
	GrafPtr port;

	GetPort(&port);
	SetPort(wind);
	GetFontInfo(&fInfo);
	SetPort(port);
	return fInfo.ascent + fInfo.descent + fInfo.leading;
}



/*----------------------------------------------------------------------------
	OutlineRegion 
	
	Change a region into a tracing of its border which is appropriate 
	for normal dragging.
			
	Entry:	theRgn = handle to region.
	
	Exit:	Region changed to outline of region.
	
	From Apple "HFS Drag Sample" sample code.
----------------------------------------------------------------------------*/

void OutlineRegion (RgnHandle theRgn)
{
	RgnHandle tempRgn;
	
	tempRgn = NewRgn();
	CopyRgn(theRgn, tempRgn);
	InsetRgn(tempRgn, 1, 1);
	DiffRgn(theRgn, tempRgn, theRgn);
	DisposeRgn(tempRgn);
}



/*----------------------------------------------------------------------------
	drawutil_InitUPP
	
	Initialize UPPs.
----------------------------------------------------------------------------*/

void drawutil_InitUPP (void)
{
	gDrawGrayRectOneDepthUPP = NewDeviceLoopDrawingProc(DrawGrayRectOneDepth);
	gDrawGrayRoundRectOneDepthUPP = NewDeviceLoopDrawingProc(DrawGrayRoundRectOneDepth);
	gDrawColorPictOneDepthUPP = NewDeviceLoopDrawingProc(DrawColorPictOneDepth);
	gFillColorPolyOneDepthUPP = NewDeviceLoopDrawingProc(FillColorPolyOneDepth);
}