/*----------------------------------------------------------------------------

	ic.c
	
	This module encapsulates the interface to the Internet Config system.
	
	Copyright  1994, Northwestern University.

----------------------------------------------------------------------------*/

#include <stdio.h>
#include <string.h>
#include <GestaltEqu.h>
#include "glob.h"
#include "ic.h"
#include "ICTypes.h"
#include "ICAPI.h"
#include "strutil.h"
#include "fileutil.h"
#include "memutil.h"
#include "apputil.h"



static ICInstance gInst;
static Boolean gICStartOK = false;



/*----------------------------------------------------------------------------
	MyICGetPref
	
	Get a preference from IC. Same parameters as ICGetPref.
----------------------------------------------------------------------------*/

static pascal ICError MyICGetPref (ICInstance inst, ConstStr255Param key, 
	ICAttr *attr, Ptr buf, long *size)
{
	Str255 keyCopy;		/* word-aligned to avoid crashes on 68000 Macs */
	
	CopyPascalString(keyCopy, (StringPtr)key);
	return ICGetPref(inst, keyCopy, attr, buf, size);
}



/*----------------------------------------------------------------------------
	MyICSetPref
	
	Set a preference to IC. Same parameters as ICSetPref.
----------------------------------------------------------------------------*/

static pascal ICError MyICSetPref (ICInstance inst, ConstStr255Param key, 
	ICAttr attr, Ptr buf, long size)
{
	Str255 keyCopy;		/* word-aligned to avoid crashes on 68000 Macs */
	
	CopyPascalString(keyCopy, (StringPtr)key);
	return ICSetPref(inst, keyCopy, attr, buf, size);
}



/*----------------------------------------------------------------------------
	MyICStart 
	
	Start Internet Config.
	
	Entry:	prefsFileLocated = true if we have found a NewsWatcher Prefs file, 
				in which case the vRefNum and dirID params are valid, 
				and we tell Internet Config to search the same folder.
			vRefNum = vol ref num of NewsWatcher Prefs file.
			dirID = dir id of NewsWatcher Prefs file.
			
	This function must be called after locating and reading NewsWatcher's 
	prefs file.
----------------------------------------------------------------------------*/

void MyICStart (Boolean prefsFileLocated, short vRefNum, long dirID)
{
	ICError icErr, icErrCountPref;
	ICDirSpecArray	dirSpecArray;
	long count;
	
	if (gICStartOK) return;
	icErr = ICStart(&gInst, kNewsWatcherSignature);
	if (icErr != noErr) return;
	if (prefsFileLocated) {
		dirSpecArray[0].vRefNum = vRefNum;
		dirSpecArray[0].dirID = dirID;
		icErr = ICFindConfigFile(gInst, 1, &dirSpecArray);
	} else {
		icErr = ICFindConfigFile(gInst, 0, nil);
	}
	if (icErr != noErr) return;
	icErr = ICBegin(gInst, icReadOnlyPerm);
	if (icErr != noErr) return;
	icErrCountPref = ICCountPref(gInst, &count);
	icErr = ICEnd(gInst);
	if (icErr != noErr) return;
	gICStartOK = icErrCountPref == noErr && count > 0;
}



/*----------------------------------------------------------------------------
	MyICStop 
	
	Stop Internet Config.
----------------------------------------------------------------------------*/

void MyICStop (void)
{
	if (gICStartOK) ICStop(gInst);
	gInst = nil;
	gICStartOK = false;
}



/*----------------------------------------------------------------------------
	ReadCStringFromPString 
	
	Read a C string shared pref from a P string IC pref.
	
	Entry:	maxSize = max size of theString.
			key = IC keyword.
			
	Exit:	theString = pref string as read from IC.
----------------------------------------------------------------------------*/

static void ReadCStringFromPString (char *theString, long maxSize, 
	ConstStr255Param key)
{
	CStr255 str;
	ICError icErr;
	ICAttr attr;
	
	icErr = MyICGetPref(gInst, key, &attr, (Ptr)str, &maxSize);
	if (icErr == noErr) {
		p2cstr((StringPtr)str);
		strcpy(theString, str);
	}
}



/*----------------------------------------------------------------------------
	ReadPStringFromPString 
	
	Read a P string shared pref from a P string IC pref.
	
	Entry:	maxSize = max size of theString.
			key = IC keyword.
			
	Exit:	theString = pref string as read from IC.
----------------------------------------------------------------------------*/

static void ReadPStringFromPString (StringPtr theString, long maxSize, 
	ConstStr255Param key)
{
	Str255 str;
	ICError icErr;
	ICAttr attr;
	
	icErr = MyICGetPref(gInst, key, &attr, (Ptr)str, &maxSize);
	if (icErr == noErr) 
		CopyPascalString(theString, str);
}



/*----------------------------------------------------------------------------
	ReadCStringFromText 
	
	Read a C string shared pref from a text IC pref.
	
	Entry:	maxSize = max size of theString.
			key = IC keyword.
			
	Exit:	theString = pref string as read from IC.
----------------------------------------------------------------------------*/

static void ReadCStringFromText (char *theString, long maxSize, 
	ConstStr255Param key)
{
	char buf[4096];
	ICError icErr;
	ICAttr attr;
	long size;
	
	size = maxSize - 1;
	icErr = MyICGetPref(gInst, key, &attr, (Ptr)buf, &size);
	if (icErr == noErr) {
		BlockMoveData(buf, theString, size);
		theString[size] = 0;
	}
}



/*----------------------------------------------------------------------------
	ReadFolderFromFileSpec 
	
	Read a default folder shared pref from an ICFileSpec IC pref.
	
	Entry:	key = IC keyword.
			
	Exit:	*defaultFolder = true if pref read OK.
			*defaultFolderAlias = handle to default folder alias.
----------------------------------------------------------------------------*/

static void ReadFolderFromFileSpec (Boolean *defaultFolder, 
	AliasHandle *defaultFolderAlias, 
	ConstStr255Param key)
{
	char buf[4096];
	ICError icErr;
	ICAttr attr;
	long size;
	ICFileSpec *p;
	AliasHandle alias;
	OSErr err = noErr;
	
	size = sizeof(buf);
	icErr = MyICGetPref(gInst, key, &attr, (Ptr)buf, &size);
	if (icErr == noErr) {
		p = (ICFileSpec*)buf;
		if (p->alias.aliasSize == 0) {
			err = VolNameAndCreationDateToVRefNum(p->vol_name, p->vol_creation_date,
				&p->fss.vRefNum);
			if (err != noErr) return;
			err = NewAlias(nil, &p->fss, &alias);
			if (err != noErr) return;
		} else {
			err = MyNewHandle(p->alias.aliasSize, &alias);
			if (err != noErr) return;
			BlockMoveData(&p->alias, *alias, p->alias.aliasSize);
		}
		MyDisposeHandle(*defaultFolderAlias);
		*defaultFolderAlias = alias;
		*defaultFolder = true;
	}
}



/*----------------------------------------------------------------------------
	ReadFontInfo 
	
	Read font info from an ICFontRecord IC pref.
	
	Entry:	key = IC keyword.
			
	Exit:	fontName = font name as read from IC.
			*fontSize = font size as read from IC.
----------------------------------------------------------------------------*/

static void ReadFontInfo (Str255 fontName, short *fontSize, 
	ConstStr255Param key)
{
	ICError icErr;
	ICFontRecord icFontRecord;
	ICAttr attr;
	long size;

	size = sizeof(icFontRecord);
	icErr = MyICGetPref(gInst, key, &attr, (Ptr)&icFontRecord, &size);
	if (icErr == noErr) {
		CopyPascalString(fontName, icFontRecord.font);
		*fontSize = icFontRecord.size;
	}
}



/*----------------------------------------------------------------------------
	ReadURLHelperInfo 
	
	Read URL Helper info from an ICAppSpec IC pref.
	
	Entry:	key = IC keyword.
			
	Exit:	*helperInfo = helper info.
----------------------------------------------------------------------------*/

static void ReadURLHelperInfo (TURLHelperInfo *helperInfo, ConstStr255Param key)
{
	ICError icErr;
	ICAppSpec icAppSpec;
	ICAttr attr;
	long size;
	
	size = sizeof(icAppSpec);
	icErr = MyICGetPref(gInst, key, &attr, (Ptr)&icAppSpec, &size);
	if (icErr == noErr && helperInfo->sig != icAppSpec.fCreator) {
		helperInfo->sig = icAppSpec.fCreator;
		helperInfo->versionNumber = 0;
		helperInfo->lastMod = 0;
	}	
}



/*----------------------------------------------------------------------------
	ReadEditorInfo 
	
	Read text editor info from an ICAppSpec IC pref.
			
	Exit:	*sig = signature of editor applicaton.
----------------------------------------------------------------------------*/

static void ReadEditorInfo (OSType *sig)
{
	ICError icErr;
	ICAppSpec icAppSpec;
	ICAttr attr;
	long size;
	
	size = sizeof(icAppSpec);
	icErr = MyICGetPref(gInst, kICeditorHelper, &attr, (Ptr)&icAppSpec, &size);
	if (icErr == noErr) *sig = icAppSpec.fCreator;
}



/*----------------------------------------------------------------------------
	MyICReadSharedPrefs 
	
	Read shared Internet Config prefs.
	
	Entry:	key = IC keyword for shared pref.
----------------------------------------------------------------------------*/

void MyICReadSharedPrefs (ConstStr255Param key)
{
	ICError icErr;

	if (!gICStartOK) return;
	
	icErr = ICBegin(gInst, icReadOnlyPerm);
	if (icErr != noErr) return;
	
	if (EqualString(key, kICAllSharedPrefs, true, true)) {
		ReadCStringFromPString(gPrefs.fullName, sizeof(gPrefs.fullName), 
			kICRealName);
		ReadCStringFromPString(gPrefs.organization, sizeof(gPrefs.organization),
			kICOrganization);
		ReadCStringFromPString(gPrefs.quoteString, sizeof(gPrefs.quoteString),
			kICQuotingString);
		ReadCStringFromText(gPrefs.signature, sizeof(gPrefs.signature),
			kICSignature);
		ReadCStringFromPString(gPrefs.emailAddress, sizeof(gPrefs.emailAddress),
			kICEmail);
		ReadPStringFromPString(gPrefs.mailServerName, sizeof(gPrefs.mailServerName),
			kICSMTPHost);
		ReadCStringFromText(gPrefs.extraMailHdrLines, sizeof(gPrefs.extraMailHdrLines),
			kICMailHeaders);
		ReadCStringFromPString(gPrefs.authUsername, sizeof(gPrefs.authUsername),
			kICNewsAuthUsername);
		ReadCStringFromPString(gPrefs.authPassword, sizeof(gPrefs.authPassword),
			kICNewsAuthPassword);
		ReadPStringFromPString(gPrefs.newsServerName, sizeof(gPrefs.newsServerName),
			kICNNTPHost);
		ReadCStringFromText(gPrefs.extraNewsHdrLines, sizeof(gPrefs.extraNewsHdrLines),
			kICNewsHeaders);
		ReadFolderFromFileSpec(&gPrefs.savedBinDefaultFolder, 
			&gPrefs.savedBinDefaultFolderAlias, kICDownloadFolder);
		ReadFontInfo(gPrefs.textFont, &gPrefs.textSize, kICScreenFont);
		ReadFontInfo(gPrefs.printingFont, &gPrefs.printingSize, kICPrinterFont);
		ReadURLHelperInfo(&gPrefs.ftpHelperInfo, kICftpHelper);
		ReadURLHelperInfo(&gPrefs.httpHelperInfo, kIChttpHelper);
		ReadURLHelperInfo(&gPrefs.gopherHelperInfo, kICgopherHelper);
		ReadURLHelperInfo(&gPrefs.waisHelperInfo, kICwaisHelper);
		ReadURLHelperInfo(&gPrefs.telnetHelperInfo, kICtelnetHelper);
		ReadURLHelperInfo(&gPrefs.tn3270HelperInfo, kICtn3270Helper);
		ReadURLHelperInfo(&gPrefs.fingerHelperInfo, kICfingerHelper);
		ReadURLHelperInfo(&gPrefs.whoisHelperInfo, kICwhoisHelper);
		ReadEditorInfo(&gPrefs.savedArtCreator);
	} else if (EqualString(key, kICRealName, true, true)) {
		ReadCStringFromPString(gPrefs.fullName, sizeof(gPrefs.fullName), 
			kICRealName);
	} else if (EqualString(key, kICOrganization, true, true)) {
		ReadCStringFromPString(gPrefs.organization, sizeof(gPrefs.organization),
			kICOrganization);
	} else if (EqualString(key, kICQuotingString, true, true)) {
		ReadCStringFromPString(gPrefs.quoteString, sizeof(gPrefs.quoteString),
			kICQuotingString);
	} else if (EqualString(key, kICSignature, true, true)) {
		ReadCStringFromText(gPrefs.signature, sizeof(gPrefs.signature),
			kICSignature);
	} else if (EqualString(key, kICEmail, true, true)) {
		ReadCStringFromPString(gPrefs.emailAddress, sizeof(gPrefs.emailAddress),
			kICEmail);
	} else if (EqualString(key, kICSMTPHost, true, true)) {
		ReadPStringFromPString(gPrefs.mailServerName, sizeof(gPrefs.mailServerName),
			kICSMTPHost);
	} else if (EqualString(key, kICMailHeaders, true, true)) {
		ReadCStringFromText(gPrefs.extraMailHdrLines, sizeof(gPrefs.extraMailHdrLines),
			kICMailHeaders);
	} else if (EqualString(key, kICNewsAuthUsername, true, true)) {
		ReadCStringFromPString(gPrefs.authUsername, sizeof(gPrefs.authUsername),
			kICNewsAuthUsername);
	} else if (EqualString(key, kICNewsAuthPassword, true, true)) {
		ReadCStringFromPString(gPrefs.authPassword, sizeof(gPrefs.authPassword),
			kICNewsAuthPassword);
	} else if (EqualString(key, kICNNTPHost, true, true)) {
		ReadPStringFromPString(gPrefs.newsServerName, sizeof(gPrefs.newsServerName),
			kICNNTPHost);
	} else if (EqualString(key, kICNewsHeaders, true, true)) {
		ReadCStringFromText(gPrefs.extraNewsHdrLines, sizeof(gPrefs.extraNewsHdrLines),
			kICNewsHeaders);
	} else if (EqualString(key, kICDownloadFolder, true, true)) {
		ReadFolderFromFileSpec(&gPrefs.savedBinDefaultFolder, 
			&gPrefs.savedBinDefaultFolderAlias, kICDownloadFolder);
	} else if (EqualString(key, kICScreenFont, true, true)) {
		ReadFontInfo(gPrefs.textFont, &gPrefs.textSize, kICScreenFont);
	} else if (EqualString(key, kICPrinterFont, true, true)) {
		ReadFontInfo(gPrefs.printingFont, &gPrefs.printingSize, kICPrinterFont);
	} else if (EqualString(key, kICftpHelper, true, true)) {
		ReadURLHelperInfo(&gPrefs.ftpHelperInfo, kICftpHelper);
	} else if (EqualString(key, kIChttpHelper, true, true)) {
		ReadURLHelperInfo(&gPrefs.httpHelperInfo, kIChttpHelper);
	} else if (EqualString(key, kICgopherHelper, true, true)) {
		ReadURLHelperInfo(&gPrefs.gopherHelperInfo, kICgopherHelper);
	} else if (EqualString(key, kICwaisHelper, true, true)) {
		ReadURLHelperInfo(&gPrefs.waisHelperInfo, kICwaisHelper);
	} else if (EqualString(key, kICtelnetHelper, true, true)) {
		ReadURLHelperInfo(&gPrefs.telnetHelperInfo, kICtelnetHelper);
	} else if (EqualString(key, kICtn3270Helper, true, true)) {
		ReadURLHelperInfo(&gPrefs.tn3270HelperInfo, kICtn3270Helper);
	} else if (EqualString(key, kICfingerHelper, true, true)) {
		ReadURLHelperInfo(&gPrefs.fingerHelperInfo, kICfingerHelper);
	} else if (EqualString(key, kICwhoisHelper, true, true)) {
		ReadURLHelperInfo(&gPrefs.whoisHelperInfo, kICwhoisHelper);
	} else if (EqualString(key, kICeditorHelper, true, true)) {
		ReadEditorInfo(&gPrefs.savedArtCreator);
	}
	
	ICEnd(gInst);
}



/*----------------------------------------------------------------------------
	WriteCStringToPString 
	
	Write a C string shared pref to a P string IC pref.
	
	Entry:	theString = pointer to C string shared pref.
			key = IC keyword.
----------------------------------------------------------------------------*/

static void WriteCStringToPString (char *theString, ConstStr255Param key)
{
	c2pstr(theString);
	MyICSetPref(gInst, key, ICattr_no_change, (Ptr)theString, *theString+1); 
	p2cstr((StringPtr)theString);
}



/*----------------------------------------------------------------------------
	WritePStringToPString 
	
	Write a P string shared pref to a P string IC pref.
	
	Entry:	theString = pointer to P string shared pref.
			key = IC keyword.
----------------------------------------------------------------------------*/

static void WritePStringToPString (StringPtr theString, ConstStr255Param key)
{
	MyICSetPref(gInst, key, ICattr_no_change, (Ptr)theString, *theString+1); 
}



/*----------------------------------------------------------------------------
	WriteCStringToText 
	
	Write a C string shared pref to a text IC pref.
	
	Entry:	theString = pointer to C string shared pref.
			key = IC keyword.
----------------------------------------------------------------------------*/

static void WriteCStringToText (char *theString, ConstStr255Param key)
{
	MyICSetPref(gInst, key, ICattr_no_change, (Ptr)theString, strlen(theString));
}



/*----------------------------------------------------------------------------
	WriteFolderToFileSpec 
	
	Write a default folder shared pref to an ICFileSpec IC pref.
	
	Entry:	defaultFolder = Boolean pref for default folder.
			defaultFolderAlias = Handle to alias for default folder
			key = IC keyword.
----------------------------------------------------------------------------*/

static void WriteFolderToFileSpec (Boolean defaultFolder, 
	AliasHandle defaultFolderAlias, 
	ConstStr255Param key)
{
	char buf[4096];
	OSErr err = noErr;
	ICFileSpec *p;
	HParamBlockRec pb;
	Str31 volName;
	long aliasSize, size;
	FSSpec fss;
	Boolean wasChanged;
	
	if (!defaultFolder) return;
	if (defaultFolderAlias == nil) return;
	err = ResolveAlias(nil, defaultFolderAlias, &fss, &wasChanged); 
	if (err != noErr) return;
	pb.volumeParam.ioNamePtr = volName;
	pb.volumeParam.ioVolIndex = 0;
	pb.volumeParam.ioVRefNum = fss.vRefNum;
	err = PBHGetVInfoSync(&pb);
	p = (ICFileSpec*)buf;
	CopyPascalString(p->vol_name, volName);
	p->vol_creation_date = pb.volumeParam.ioVCrDate;
	p->fss = fss;
	aliasSize = (**defaultFolderAlias).aliasSize;
	size = aliasSize + sizeof(Str31) + sizeof(long) + sizeof(FSSpec);
	if (size > sizeof(buf)) return;
	BlockMoveData(*defaultFolderAlias, &p->alias, aliasSize);
	MyICSetPref(gInst, key, ICattr_no_change, (Ptr)buf, size);
}



/*----------------------------------------------------------------------------
	WriteFontInfo 
	
	Read font info to an ICFontRecord IC pref.
	
	Entry:	fontName = font name.
			fontSize = font size.
			key = IC keyword.
----------------------------------------------------------------------------*/

static void WriteFontInfo (Str255 fontName, short fontSize, 
	ConstStr255Param key)
{
	ICError icErr;
	ICFontRecord icFontRecord;
	ICAttr attr;
	long size;

	size = sizeof(icFontRecord);
	icErr = MyICGetPref(gInst, key, &attr, (Ptr)&icFontRecord, &size);
	if (icErr == icPrefNotFoundErr) {
		icFontRecord.face = normal;
	} else if (icErr != noErr) {
		return;
	}
	CopyPascalString(icFontRecord.font, fontName);
	icFontRecord.size = fontSize;
	MyICSetPref(gInst, key, ICattr_no_change, (Ptr)&icFontRecord, size);
}



/*----------------------------------------------------------------------------
	WriteURLHelperInfo 
	
	Write URL Helper info to an ICAppSpec IC pref.
	
	Entry:	helperInfo = pointer to helper info.
			key = IC keyword.
----------------------------------------------------------------------------*/

static void WriteURLHelperInfo (TURLHelperInfo *helperInfo, ConstStr255Param key)
{
	ICAppSpec icAppSpec;
	long size;
	OSErr err = noErr;
	
	icAppSpec.fCreator = helperInfo->sig;
	err = FindAppNameFromSig(icAppSpec.fCreator, icAppSpec.name);
	if (err != noErr) *icAppSpec.name = 0;
	size = sizeof(icAppSpec);
	MyICSetPref(gInst, key, ICattr_no_change, (Ptr)&icAppSpec, size);
}



/*----------------------------------------------------------------------------
	WriteEditorInfo 
	
	Write Editor info to the "editor" helper ICAppSpec IC pref.
	
	Entry:	sig = editor helper signature.
----------------------------------------------------------------------------*/

static void WriteEditorInfo (OSType sig)
{
	ICAppSpec icAppSpec;
	long size;
	OSErr err = noErr;
	
	icAppSpec.fCreator = sig;
	err = FindAppNameFromSig(sig, icAppSpec.name);
	if (err != noErr) *icAppSpec.name = 0;
	size = sizeof(icAppSpec);
	MyICSetPref(gInst, kICeditorHelper, ICattr_no_change, (Ptr)&icAppSpec, size);
}



/*----------------------------------------------------------------------------
	MyICWriteSharedPrefs 
	
	Write shared Internet Config prefs.
	
	Entry:	key = IC keyword for shared pref.
----------------------------------------------------------------------------*/

void MyICWriteSharedPrefs (ConstStr255Param key)
{
	ICError icErr;

	if (!gICStartOK) return;
	
	icErr = ICBegin(gInst, icReadWritePerm);
	if (icErr != noErr) return;
	
	if (EqualString(key, kICAllSharedPrefs, true, true)) {
		WriteCStringToPString(gPrefs.fullName, kICRealName);
		WriteCStringToPString(gPrefs.organization, kICOrganization);
		WriteCStringToPString(gPrefs.quoteString, kICQuotingString);
		WriteCStringToText(gPrefs.signature, kICSignature);
		WriteCStringToPString(gPrefs.emailAddress, kICEmail);
		WritePStringToPString(gPrefs.mailServerName, kICSMTPHost);
		WriteCStringToText(gPrefs.extraMailHdrLines, kICMailHeaders);
		WriteCStringToPString(gPrefs.authUsername, kICNewsAuthUsername);
		WriteCStringToPString(gPrefs.authPassword, kICNewsAuthPassword);
		WritePStringToPString(gPrefs.newsServerName, kICNNTPHost);
		WriteCStringToText(gPrefs.extraNewsHdrLines, kICNewsHeaders);
		WriteFolderToFileSpec(gPrefs.savedBinDefaultFolder,
			gPrefs.savedBinDefaultFolderAlias, kICDownloadFolder); 
		WriteFontInfo(gPrefs.textFont, gPrefs.textSize, kICScreenFont);
		WriteFontInfo(gPrefs.printingFont, gPrefs.printingSize, kICPrinterFont);
		WriteURLHelperInfo(&gPrefs.ftpHelperInfo, kICftpHelper);
		WriteURLHelperInfo(&gPrefs.httpHelperInfo, kIChttpHelper);
		WriteURLHelperInfo(&gPrefs.gopherHelperInfo, kICgopherHelper);
		WriteURLHelperInfo(&gPrefs.waisHelperInfo, kICwaisHelper);
		WriteURLHelperInfo(&gPrefs.telnetHelperInfo, kICtelnetHelper);
		WriteURLHelperInfo(&gPrefs.tn3270HelperInfo, kICtn3270Helper);
		WriteURLHelperInfo(&gPrefs.fingerHelperInfo, kICfingerHelper);
		WriteURLHelperInfo(&gPrefs.whoisHelperInfo, kICwhoisHelper);
		WriteEditorInfo(gPrefs.savedArtCreator);
	} else if (EqualString(key, kICRealName, true, true)) {
		WriteCStringToPString(gPrefs.fullName, kICRealName);
	} else if (EqualString(key, kICOrganization, true, true)) {
		WriteCStringToPString(gPrefs.organization, kICOrganization);
	} else if (EqualString(key, kICQuotingString, true, true)) {
		WriteCStringToPString(gPrefs.quoteString, kICQuotingString);
	} else if (EqualString(key, kICSignature, true, true)) {
		WriteCStringToText(gPrefs.signature, kICSignature);
	} else if (EqualString(key, kICEmail, true, true)) {
		WriteCStringToPString(gPrefs.emailAddress, kICEmail);
	} else if (EqualString(key, kICSMTPHost, true, true)) {
		WritePStringToPString(gPrefs.mailServerName, kICSMTPHost);
	} else if (EqualString(key, kICMailHeaders, true, true)) {
		WriteCStringToText(gPrefs.extraMailHdrLines, kICMailHeaders);
	} else if (EqualString(key, kICNewsAuthUsername, true, true)) {
		WriteCStringToPString(gPrefs.authUsername, kICNewsAuthUsername);
	} else if (EqualString(key, kICNewsAuthPassword, true, true)) {
		WriteCStringToPString(gPrefs.authPassword, kICNewsAuthPassword);
	} else if (EqualString(key, kICNNTPHost, true, true)) {
		WritePStringToPString(gPrefs.newsServerName, kICNNTPHost);
	} else if (EqualString(key, kICNewsHeaders, true, true)) {
		WriteCStringToText(gPrefs.extraNewsHdrLines, kICNewsHeaders);
	} else if (EqualString(key, kICDownloadFolder, true, true)) {
		WriteFolderToFileSpec(gPrefs.savedBinDefaultFolder,
			gPrefs.savedBinDefaultFolderAlias, kICDownloadFolder); 
	} else if (EqualString(key, kICScreenFont, true, true)) {
		WriteFontInfo(gPrefs.textFont, gPrefs.textSize, kICScreenFont);
	} else if (EqualString(key, kICPrinterFont, true, true)) {
		WriteFontInfo(gPrefs.printingFont, gPrefs.printingSize, kICPrinterFont);
	} else if (EqualString(key, kICftpHelper, true, true)) {
		WriteURLHelperInfo(&gPrefs.ftpHelperInfo, kICftpHelper);
	} else if (EqualString(key, kIChttpHelper, true, true)) {
		WriteURLHelperInfo(&gPrefs.httpHelperInfo, kIChttpHelper);
	} else if (EqualString(key, kICgopherHelper, true, true)) {
		WriteURLHelperInfo(&gPrefs.gopherHelperInfo, kICgopherHelper);
	} else if (EqualString(key, kICwaisHelper, true, true)) {
		WriteURLHelperInfo(&gPrefs.waisHelperInfo, kICwaisHelper);
	} else if (EqualString(key, kICtelnetHelper, true, true)) {
		WriteURLHelperInfo(&gPrefs.telnetHelperInfo, kICtelnetHelper);
	} else if (EqualString(key, kICtn3270Helper, true, true)) {
		WriteURLHelperInfo(&gPrefs.tn3270HelperInfo, kICtn3270Helper);
	} else if (EqualString(key, kICfingerHelper, true, true)) {
		WriteURLHelperInfo(&gPrefs.fingerHelperInfo, kICfingerHelper);
	} else if (EqualString(key, kICwhoisHelper, true, true)) {
		WriteURLHelperInfo(&gPrefs.whoisHelperInfo, kICwhoisHelper);
	} else if (EqualString(key, kICeditorHelper, true, true)) {
		WriteEditorInfo(gPrefs.savedArtCreator);
	}
	
	ICEnd(gInst);
}
