/*----------------------------------------------------------------------------

	nntp.c

	This reusable module implements an interface to NNTP (USENET news) servers.
	
	The following functions are exported:
	
		NntpOpen - Open an NNTP stream.
		NntpClose - Close an NNTP stream.
		NntpAbort - Abort an NNTP stream.
		NntpIdle - Perform idle time tasks.
		NntpSetStreamOptions - Set stream options.
		NntpGetGroupList - Get full group list or new groups list.
		NntpGetGroupNames - Get group names.
		NntpGetGroupInfo - Get info for a single group.
		NntpGetMultipleGroupInfo - Get info for multiple groups.
		NntpGetHeaders - Get selected article headers.
		NntpGetArticle - Get an article.
		NntpGetChunkyArticle - Get an article in chunks.
		NntpPostArticle - Post an article.
		NntpAuthorize - Authorize user.
		NntpGetHello - Get server hello message.
		NntpGetHelp - Get server help text.
		NntpGetIPAddr - Get server IP address.
		
	The following functions are used to get error information:
		
		NntpGetServerErrInfo - Get server error information.
		NntpGetOSErr - Get OS error number.
		
	The following utility functions are also exported:
	
		IsLegalBinHexLine - Check line for BinHex encoded text.
		IsLegalUULine - Check line for uuencode text.
		
	The NetInit function in module net.c must be called before calling any of
	the functions in this module. You also must call both the NntpIdle and the 
	NetIdle functions in your idle loop, and the NetTerm function at program termination.
		
	A "stream" is an abstraction representing a bidirectional network connection
	to an NNTP server. A stream is represented as a handle of type "NntpStreamHandle". 
	These stream handles are opaque. You may copy them and pass them as parameters 
	to functions in nntp.c, but you are prohibited from accessing the contents of
	the memory blocks referenced by the stream handles. Only the functions in 
	nntp.c are permitted to manipulate the contents of these blocks.
	
	The functions return a value of type NntpErr as the function result:
	
		nntpNoErr				no error occurred
		nntpNoSuchGroupErr		no such group or group access restriction
		nntpNoSuchArticleErr	no such article
		nntpServerErr			server error
		nntpOSErr				OS error
		
	If the function result is nntpServerErr, the NntpGetServerErrInfo function should
	be called to get information about the server error. On server errors, no such group
	errors, and no such article errors, the stream is still open on return to the caller
	(except in NntpOpen, which creates the stream only if no errors of any kind occur).
	
	If the function result is nntpOSErr, the NntpGetOSErr function should be called
	to get the OS error number:

		userCanceledErr		if the user canceled the operation
		other				any other OS, Toolbox, or MacTCP error code
		
	On an OS error (which includes userCanceledErr), the connection to the
	news server may be aborted, but the NNTP stream is left allocated. The module will
	automatically attempt to establish a new connection the next time the stream is
	used.
	
	All group names, header names, message id, and pattern strings passed as parameters
	to functions in this module are restricted to a maximum of 127 characters. If they 
	exceed this maximum, they are truncated.
	
	The following options are associated with each stream:
	
		idleTime		stream is automatically closed if idle for this 
						many minutes, or 0 for no automatic close
		useXPAT			true to use XPAT command for searches
		sendModeReader	true to send MODE READER command after connect
		batchedCmds		true to use batched GROUP commands
		newConnection	true to establish new connection before getting
						group info
		
	The default values are idleTime=0, useXPAT=false, sendModeReader=true, 
	batchedCmds=false, and newConnection=true. These options should work with
	any properly functioning NNTP server. The four Boolean options can be changed
	to improve performance, but they don't work with all servers. See the individual
	function descriptions more for more details.
	
	Copyright  1994, Northwestern University.

----------------------------------------------------------------------------*/

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <ctype.h>

#include "def.h"
#include "nntp.h"
#include "net.h"
#include "strutil.h"
#include "qsort.h"
#include "memutil.h"



/* Types. */

typedef struct TStream {
	NetStreamHandle netStream;			/* net stream handle, or nil if closed */
	NntpStreamOptions options;			/* stream options */
	Boolean serverHasXPAT;				/* true if server supports XPAT command */
	unsigned long addr;					/* server IP address */
	unsigned short port;				/* server port number */
	CStr255 helloMsg;					/* hello message */
	char curGroup[128];					/* current group on server */
	unsigned long lastTransactionTime;	/* time of last transaction on stream */
	struct TStream **next;				/* handle to next open NNTP stream */
} TStream, *TStreamPtr, **TStreamHandle;
	


/*	Global variables. */

static OSErr gErr = noErr;				/* OS error code */
static CStr255 gCommand = "";			/* server command */
static long gResponseCode = 0;			/* server response code */
static CStr255 gResponse = "";			/* server response message */
static TStreamHandle gStreamList = nil;	/* handle to list of open NNTP streams */
static Boolean gRequireEncodedTextBeginLine;	/* true if attached file must
												   include "begin" flag line */

/*	Globals used by DoOneGroupCmdResponse and NntpGetMultipleGroupInfo. */

static NntpGroupInfoHandle gInfo;		/* handle to group info array */
static long gInfoIndex;					/* current index in group info array */
static Boolean gGroupInfoServerError;	/* true if server error encountered */



/*----------------------------------------------------------------------------
	SetLastTransactionTime 
	
	Record the transaction time on an NNTP stream.
	
	Entry:	s = handle to stream.
----------------------------------------------------------------------------*/

static void SetLastTransactionTime (TStreamHandle s)
{
	unsigned long lastTransactionTime;
	
	GetDateTime(&lastTransactionTime);
	(**s).lastTransactionTime = lastTransactionTime;
}



/*----------------------------------------------------------------------------
	SetCurrentGroup 
	
	Record the current group on an NNTP stream.
	
	Entry:	s = handle to stream.
			group = group name.
----------------------------------------------------------------------------*/

static void SetCurrentGroup (TStreamHandle s, char *group)
{
	short len;
	
	len = strlen(group);
	if (len > 127) len = 127;
	BlockMoveData(group, (**s).curGroup, len);
	*((**s).curGroup + len) = 0;
}



/*----------------------------------------------------------------------------
	Authorize 
	
	Authorize user.
	
	Entry:	netStream = pointer to net stream.
			username = username.
			password = password.
			
	Exit:	function result = result code.
----------------------------------------------------------------------------*/

static NntpErr Authorize (NetStreamHandle netStream, char *username, char *password)
{
	if (*username == 0 || *password == 0) return nntpNoErr;

	sprintf(gCommand, "AUTHINFO USER %s", username);
	gErr = NetCommand(netStream, gCommand, &gResponseCode, gResponse);
	if (gErr != noErr) goto exit1;
	if (gResponseCode != 381) goto exit2;
	
	sprintf(gCommand, "AUTHINFO PASS %s", password);
	gErr = NetCommand(netStream, gCommand, &gResponseCode, gResponse);
	if (gErr != noErr) goto exit1;
	if (gResponseCode != 281) goto exit2;
	
	return nntpNoErr;
	
exit1:

	return nntpOSErr;
	
exit2:

	NetClose(netStream);
	return nntpServerErr;
}



/*----------------------------------------------------------------------------
	ReOpenConnection 
	
	Reopen a connection to an NNTP server.
	
	Entry:	s = handle to stream.
	
	Exit:	function result = result code.
----------------------------------------------------------------------------*/

static NntpErr ReOpenConnection (TStreamHandle s)
{
	NetStreamHandle netStream;
	char username[32];
	char password[32];
	NntpErr nntpErr;
	
	if ((**s).netStream != nil) return nntpNoErr;
	
	NetIdle();
	
	gErr = NetOpen((**s).addr, (**s).port, &netStream, &gResponseCode, gResponse);
	if (gErr != noErr) return nntpOSErr;
	if (gResponseCode != 200 && gResponseCode != 201) return nntpServerErr;
	strcpy((**s).helloMsg, gResponse);

	if ((**s).options.sendModeReader) {
		strcpy(gCommand, "MODE READER");
		gErr = NetCommand(netStream, gCommand, &gResponseCode, gResponse);
		if (gErr != noErr) return nntpOSErr;
	}
	
	if ((**s).options.authOnConnect) {
		strcpy(username, (**s).options.username);
		strcpy(password, (**s).options.password);
		nntpErr = Authorize(netStream, username, password);
		if (nntpErr != nntpNoErr) goto exit;
	}
	
	(**s).netStream = netStream;
	*(**s).curGroup = 0;
	SetLastTransactionTime(s);

	return nntpNoErr;
	
exit:

	(**s).netStream = nil;
	return nntpErr;
}


/*----------------------------------------------------------------------------
	DoOneGroupCmdResponse 
	
	Process one GROUP command response for batched GROUP commands.
	
	Entry:	responseCode = resonse code.
			response = response string.
			
	This function is used by NntpGetMultipleGroupInfo.
----------------------------------------------------------------------------*/

static void DoOneGroupCmdResponse (long responseCode, CStr255 response)
{
	NntpGroupInfoPtr x;
	char *p;
	char state;

	if (gGroupInfoServerError) return;
	state = MyHGetState(gInfo);
	MyHLock(gInfo);
	x = *gInfo + gInfoIndex;
	if (responseCode == 211) {
		p = response;
		CrackNum(&p);
		x->count = CrackNum(&p);
		x->first = CrackNum(&p);
		x->last = CrackNum(&p);
		x->ok = true;
	} else if (responseCode == 502) {
		x->count = 0;
		x->first = 1;
		x->last = 0;
		x->ok = true;
	} else if (responseCode == 411) {
		x->ok = false;
	} else {
		gGroupInfoServerError = true;
	}
	gInfoIndex++;
	MyHSetState(gInfo, state);
}



/*----------------------------------------------------------------------------
	SortHeadersCompare 
	
	This comparison function used to sort a header info array into increasing 
	order by article number.
	
	Entry:	p = pointer to first NntpHeaderInfo struct.
			q = pointer to second NntpHeaderInfo struct.
			
	Exit:	function result =
				-1 if p->number < q->number
				 0 if p->number == q->number
				 1 if p->number > q->number
----------------------------------------------------------------------------*/

static short SortHeadersCompare (NntpHeaderInfoPtr p, NntpHeaderInfoPtr q)
{
	static short counter = 0;

	if((++counter & 0x1f) == 0) {
		if (NetGiveTime(false) == userCanceledErr) CancelFastQSort();
		counter = 0;
	}
	
	if (p->number < q->number) {
		return -1;
	} else if (p->number == q->number) {
		return 0;
	} else {
		return 1;
	}
}



/*----------------------------------------------------------------------------
	IsLegalBinHexLine 
	
	Check a line to see if it contains legal BinHex encoded text.
	
	Entry:	p = pointer to line.
			len = length of line.
			
	Exit:	function result = true if legal BinHex.
----------------------------------------------------------------------------*/

Boolean IsLegalBinHexLine (unsigned char *p, long len)
{
	unsigned char *pEnd, *pBegin;
	unsigned char c;
	
	static Boolean legalBinHexChar[] = {
		/* 0x00 */
			false,	false,	false,	false,	false,	false,	false,	false,
			false,	false,	false,	false,	false,	false,	false,	false,
		/* 0x10 */
			false,	false,	false,	false,	false,	false,	false,	false,
			false,	false,	false,	false,	false,	false,	false,	false,
		/* 0x20 */
			false,	true,	true,	true,	true,	true,	true,	true,
			true,	true,	true,	true,	true,	true,	false,	false,
		/* 0x30 */
			true,	true,	true,	true,	true,	true,	true,	false,
			true,	true,	true,	false,	false,	false,	false,	false,
		/* 0x40 */
			true,	true,	true,	true,	true,	true,	true,	true,
			true,	true,	true,	true,	true,	true,	true,	false,
		/* 0x50 */
			true,	true,	true,	true,	true,	true,	true,	false,
			true,	true,	true,	true,	false,	false,	false,	false,
		/* 0x60 */
			true,	true,	true,	true,	true,	true,	true,	false,
			true,	true,	true,	true,	true,	true,	false,	false,
		/* 0x70 */
			true,	true,	true,	false,	false,	false,	false,	false,
			false,	false,	false,	false,	false,	false,	false,	false,
		/* 0x80 */
			false,	false,	false,	false,	false,	false,	false,	false,
			false,	false,	false,	false,	false,	false,	false,	false,
			false,	false,	false,	false,	false,	false,	false,	false,
			false,	false,	false,	false,	false,	false,	false,	false,
			false,	false,	false,	false,	false,	false,	false,	false,
			false,	false,	false,	false,	false,	false,	false,	false,
			false,	false,	false,	false,	false,	false,	false,	false,
			false,	false,	false,	false,	false,	false,	false,	false,
			false,	false,	false,	false,	false,	false,	false,	false,
			false,	false,	false,	false,	false,	false,	false,	false,
			false,	false,	false,	false,	false,	false,	false,	false,
			false,	false,	false,	false,	false,	false,	false,	false,
			false,	false,	false,	false,	false,	false,	false,	false,
			false,	false,	false,	false,	false,	false,	false,	false,
			false,	false,	false,	false,	false,	false,	false,	false,
			false,	false,	false,	false,	false,	false,	false,	false,
		};

	for (pBegin = p, pEnd = p + len; p < pEnd; p++)
		if (!legalBinHexChar[*p]) return false;
	c = *pBegin;
	for (p = pBegin+1; p < pEnd; p++)
		if (*p != c) return true;
	return false;
}



/*----------------------------------------------------------------------------
	IsLegalUULine 
	
	Check a line to see if it contains legal uuencode text.
	
	Entry:	p = pointer to line.
			len = length of line.
			
	Exit:	function result = true if legal uuencode.
----------------------------------------------------------------------------*/

Boolean IsLegalUULine (unsigned char *p, long len)
{
	long n;
	unsigned char *pEnd, *pBegin;
	unsigned char c;
	
	n = *p - ' ';
	if (n < 0 || n > 63) return false;
	n = (n+2)/3*4;
	if (len > n+3 || len < n-3) return false;
	for (pBegin = p, pEnd = p + len, p++; p < pEnd; p++)
		if (*p < ' ' || *p > '`') return false;
	c = *pBegin;
	for (p = pBegin+1; p < pEnd; p++)
		if (*p != c) return true;
	return false;
}



/*----------------------------------------------------------------------------
	CheckForAttachedFile 
	
	Check for attached file and truncate article if BinHex or uuencode
	text encountered.
	
	Entry:	t = handle to raw text received from server so far.
			tLen = length of text received from server so far.
			*position = saved position in text (0 on first call).
			gRequireEncodedTextBeginLine = true if the encoded text
				must include a begin "flag" line.
			
	Exit:	function result = true if text should be truncated.
			*position = updated saved position in text if no truncation,
				length of truncated text if truncated.
				
	This function looks for the flag line (if required), followed by 
	two adjacent lines of the same length at least 60 characters long,
	each of which is either a legal BinHex line or a legal uuencode line.
----------------------------------------------------------------------------*/

static Boolean CheckForAttachedFile (Handle t, long tLen, long *position)
{
	long lenA, lenB;
	unsigned char *p, *pEnd, *q, *r;
	unsigned char *flagLine, *lineA, *lineB, *nextLine;
	Boolean haveBeginFlagLine, beginFlagLineIsBinHex, foundIt;

	p = (unsigned char*)(*t + *position);
	pEnd = (unsigned char*)(*t + tLen - 1);
	
	while (p < pEnd) {
		if (gRequireEncodedTextBeginLine) {
			haveBeginFlagLine = false;
			flagLine = p;
			while (flagLine < pEnd) {
				q = flagLine;
				while (q < pEnd && *q != CR) q++;
				while (q < pEnd && (*q == CR || *q == LF)) q++;
				if (q >= pEnd) break;
				if (*flagLine == '(') {
					if (flagLine + 39 >= pEnd) break;
					if (strncmp((char*)flagLine, 
						"(This file must be converted with BinHex", 39) == 0) 
					{
						haveBeginFlagLine = true;
						beginFlagLineIsBinHex = true;
						break;
					}
				} else if (*flagLine == 'b') {
					if (flagLine + 9 >= pEnd) break;
					if (strncmp((char*)flagLine, "begin ", 6) == 0) {
						r = flagLine + 6;
						if (isoctal(*r) && isoctal(*(r+1)) && isoctal(*(r+2))) {
							haveBeginFlagLine = true;
							beginFlagLineIsBinHex = false;
							break;
						}
					}
				} 
				flagLine = q;
			}
			if (!haveBeginFlagLine) {
				*position = (char*)flagLine - *t;
				return false;
			}
			p = flagLine;
			lineA = q;
			if (lineA >= pEnd) break;
		} else {
			lineA = p;
		}
		q = lineA;
		while (q < pEnd && *q != CR) q++;
		lenA = q - lineA;
		while (q < pEnd && (*q == CR || *q == LF)) q++;
		lineB = q;
		if (lineB >= pEnd) break;
		nextLine = gRequireEncodedTextBeginLine ? lineA : lineB;
		if (lenA < 60) {
			p = nextLine;
			continue;
		}
		q = lineB;
		while (q < pEnd && *q != CR) q++;
		if (q >= pEnd) break;
		lenB = q - lineB;
		if (lenA != lenB) {
			p = nextLine;
			continue;
		}
		if (gRequireEncodedTextBeginLine) {
			if (beginFlagLineIsBinHex) {
				foundIt = IsLegalBinHexLine(lineA, lenA) && IsLegalBinHexLine(lineB, lenB);
			} else {
				foundIt = IsLegalUULine(lineA, lenA) && IsLegalUULine(lineB, lenB);
			}
		} else {
			foundIt = (IsLegalBinHexLine(lineA, lenA) && IsLegalBinHexLine(lineB, lenB)) ||
				(IsLegalUULine(lineA, lenA) && IsLegalUULine(lineB, lenB));
		}
		if (foundIt) {
			*position = (char*)lineA - *t;
			return true;
		} else {
			p = nextLine;
			continue;
		}
	}
	
	*position = (char*)p - *t;
	return false;
}



/*----------------------------------------------------------------------------
	NntpOpen 
	
	Open an NNTP stream.
	
	Entry:	host = server host address (domain name or dotted 
				decimal IP address).
			options = pointer to server options, or nil to set default options.
	
	Exit:	function result = result code.
			stream = handle to opened stream, 
				or nil if function result != nntpNoErr.
			
	If the sendModeReader stream option is set, a "MODE READER" command is
	sent to the server after connecting. This command is also sent in the
	future on any automatic reconnects to the server. Some INN servers require
	this command.
----------------------------------------------------------------------------*/

NntpErr NntpOpen (char *host, NntpStreamOptions *options, NntpStreamHandle *stream)
{
	TStreamHandle s;
	unsigned long addr;
	unsigned short port;
	NetStreamHandle netStream;
	CStr255 helloMsg;
	NntpErr nntpErr;
	
	*stream = nil;
	*gCommand = 0;
	
	gErr = NetNameToAddr(host, kNNTPPort, &addr, &port);
	if (gErr != noErr) return nntpOSErr;
	
	gErr = NetOpen(addr, port, &netStream, &gResponseCode, gResponse);
	if (gErr != noErr) return nntpOSErr;
	if (gResponseCode != 200 && gResponseCode != 201) goto exit1;
	strcpy(helloMsg, gResponse);
	
	if (options == nil || options->sendModeReader) {
		strcpy(gCommand, "MODE READER");
		gErr = NetCommand(netStream, gCommand, &gResponseCode, gResponse);
		if (gErr != noErr) return nntpOSErr;
	}
	
	if (options != nil && options->authOnConnect) {
		nntpErr = Authorize(netStream, options->username, options->password);
		if (nntpErr != nntpNoErr) return nntpErr;
	}
	
	gErr = MyNewHandle(sizeof(TStream), &s);
	if (gErr != noErr) goto exit2;
	
	(**s).netStream = netStream;
	if (options == nil) {
		(**s).options.idleTime = 0;
		(**s).options.useXPAT = false;
		(**s).options.sendModeReader = true;
		(**s).options.batchedCmds = false;
		(**s).options.newConnection = true;
		(**s).options.authOnConnect = false;
	} else {
		(**s).options = *options;
	}
	(**s).serverHasXPAT = true;
	(**s).addr = addr;
	(**s).port = port;
	strcpy((**s).helloMsg, helloMsg);
	*(**s).curGroup = 0;
	SetLastTransactionTime(s);
	(**s).next = gStreamList;
	gStreamList = s;

	*stream = (NntpStreamHandle)s;
	return nntpNoErr;
	
exit1:

	NetClose(netStream);
	return nntpServerErr;
	
exit2:

	NetClose(netStream);
	return nntpOSErr;
	
}



/*----------------------------------------------------------------------------
	NntpClose 
	
	Close an NNTP stream.
	
	Entry:	stream = handle to stream.
	
	Exit:	function result = result code (always nntpNoErr).
----------------------------------------------------------------------------*/

NntpErr NntpClose (NntpStreamHandle stream)
{
	TStreamHandle s, prev, cur;
	NetStreamHandle netStream;
	
	s = (TStreamHandle)stream;
	
	for (prev = nil, cur = gStreamList; 
		cur != nil && cur != s; 
		prev = cur, cur = (**cur).next) /* do nothing */ ;
	if (cur != nil) {
		if (prev == nil) {
			gStreamList = (**cur).next;
		} else {
			(**prev).next = (**cur).next;
		}
	}
	
	netStream = (**s).netStream;
	if (netStream != nil) NetClose(netStream);
	MyDisposeHandle(s);
	return nntpNoErr;
}



/*----------------------------------------------------------------------------
	NntpAbort 
	
	Abort an NNTP stream.
	
	Entry:	stream = handle to stream.
	
	Exit:	function result = result code (always nntpNoErr).
----------------------------------------------------------------------------*/

NntpErr NntpAbort (NntpStreamHandle stream)
{
	TStreamHandle s;
	NetStreamHandle netStream;
	
	s = (TStreamHandle)stream;
	netStream = (**s).netStream;
	if (netStream == nil) return nntpNoErr;
	NetClose(netStream);
	(**s).netStream = nil;
	return nntpNoErr;
}



/*----------------------------------------------------------------------------
	NntpIdle 
	
	Handle idle time tasks.
	
	Exit:	function result = result code (always nntpNoErr).
	
	This function checks all the open NNTP streams to see if their idle
	timers have expired. Any streams with expired timers are closed.
----------------------------------------------------------------------------*/

NntpErr NntpIdle (void)
{
	TStreamHandle s;
	NetStreamHandle netStream;
	short idleTime;
	unsigned long lastTransactionTime, curTime;
	
	GetDateTime(&curTime);
	for (s = gStreamList; s != nil; s = (**s).next) {
		netStream = (**s).netStream;
		idleTime = (**s).options.idleTime;
		if (netStream != nil && idleTime != 0) {
			lastTransactionTime = (**s).lastTransactionTime;
			if (curTime > lastTransactionTime + 60*idleTime) {
				NetClose(netStream);
				(**s).netStream = nil;
			}
		}
	}
	return nntpNoErr;
}



/*----------------------------------------------------------------------------
	NntpSetStreamOptions
	
	Set stream options.
	
	Entry:	stream = handle to stream.
			options = pointer to stream options.
	
	Exit:	function result = result code (always nntpNoErr).
----------------------------------------------------------------------------*/

NntpErr NntpSetStreamOptions (NntpStreamHandle stream, NntpStreamOptions *options)
{
	TStreamHandle s;
	
	s = (TStreamHandle)stream;
	(**s).options = *options;
	return nntpNoErr;
}



/*----------------------------------------------------------------------------
	NntpGetGroupList 
	
	Get the full group list or a new groups list from the server.
	
	Entry:	stream = handle to stream.
			time = 0 to get full group list.
			time != 0 to only get new groups created since the specified time.
			needNumbers = true to request first, last, and count fields.
	
	Exit:	function result = result code.
			info = handle to array of group info.
			strings = handle to group name strings.
			numGroups = number of groups.
			
	The time parameter is in GetDateTime format (seconds since 01/01/04). It
	is interpreted by the server, not by the Mac. To avoid missing new groups
	due to differences between the server's time and the Mac's time, you
	should subtract a healthy amount from the time of your last new groups 
	check (e.g., 36 hours), then discard any duplicate groups returned.  
			
	The strings block contains the C-format group names. Group names are 
	truncated to 127 characters.
	
	The first, last, and status fields are read from the server.
	The count fields are computed as max(0, last - first + 1).
	The ok fields are set to true.
	
	If you don't need the first, last, and count fields, pass 
	needNumbers=false. This speeds up the function quite a bit, especially
	when getting a large full group list.
----------------------------------------------------------------------------*/

NntpErr NntpGetGroupList (NntpStreamHandle stream, unsigned long time, 
	Boolean needNumbers, NntpGroupInfoHandle *info, Handle *strings, long *numGroups)
{
	TStreamHandle s;
	NetStreamHandle netStream;
	DateTimeRec timeRec;
	Handle text=nil;
	NntpGroupInfoHandle theInfo=nil;
	char *p, *pEnd, *q, *r;
	NntpGroupInfoPtr x, xEnd;
	long n, len;
	NntpErr nntpErr;
	
	s = (TStreamHandle)stream;
	nntpErr = ReOpenConnection(s);
	if (nntpErr != nntpNoErr) return nntpErr;
	netStream = (**s).netStream;
	
	if (time == 0) {
		strcpy(gCommand, "LIST");
	} else {
		SecondsToDate(time, &timeRec);
		timeRec.year = timeRec.year % 100;
		sprintf(gCommand, "NEWGROUPS %02d%02d%02d %02d%02d%02d", 
			timeRec.year, timeRec.month, timeRec.day,
			timeRec.hour, timeRec.minute, timeRec.second);
	}
	
	gErr = NetCommand(netStream, gCommand, &gResponseCode, gResponse);
	if (gErr != noErr) goto exit1;
	if (gResponseCode != (time == 0 ? 215 : 231)) goto exit2;
	
	gErr = NetGetText(netStream, &text);
	if (gErr != noErr) goto exit1;

	pEnd = *text + MyGetHandleSize(text);
	n = 0;
	for (p = *text; p < pEnd; p++) if (*p == CR) n++;

	gErr = MyNewHandle(n*sizeof(NntpGroupInfo), &theInfo);
	if (gErr != noErr) goto exit3;

	MyHLock(text);
	MyHLock(theInfo);
	p = q = *text;
	for (x = *theInfo, xEnd = x + n; x < xEnd; x++) {
		gErr = NetGiveTime(false);
		if (gErr != noErr) goto exit3;
		r = p;
		while (*r != ' ' && *r != CR) r++;
		len = r-p;
		if (len > 127) len = 127;
		x->offset = q - *text;
		if (needNumbers) {
			x->last = CrackNum(&r);
			x->first = CrackNum(&r);
			x->count = x->last - x->first + 1;
			if (x->count < 0) x->count = 0;
		} else {
			while (*r == ' ') r++;
			while (*r != ' ' && *r != CR) r++;
			while (*r == ' ') r++;
			while (*r != ' ' && *r != CR) r++;
		}
		while (*r == ' ') r++;
		x->status = tolower(*r);
		while (*r != CR) r++;
		x->ok = true;
		*(p+len) = 0;
		strcpy(q, p);
		q += len+1;
		p = r+1;
	}
	len = q-*text;
	MyHUnlock(text);
	MyHUnlock(theInfo);
	
	MySetHandleSize(text, len);
	*info = theInfo;
	*strings = text;
	*numGroups = n;
	SetLastTransactionTime(s);

	return nntpNoErr;
	
exit1:

	MyDisposeHandle(text);
	MyDisposeHandle(theInfo);
	(**s).netStream = nil;
	return nntpOSErr;
	
exit2:
	
	MyDisposeHandle(text);
	MyDisposeHandle(theInfo);
	return nntpServerErr;
	
exit3:

	MyDisposeHandle(text);
	MyDisposeHandle(theInfo);
	return nntpOSErr;
}



/*----------------------------------------------------------------------------
	NntpGetGroupNames 
	
	Get the full list of group names or a list of names for new groups 
	from the server.
	
	Entry:	stream = handle to stream.
			time = 0 to get full group list.
			time != 0 to only get new groups created since the specified time.
			statusFilter = a C-format string of characters. Only the names of 
				groups with status not equal to one of the characters in this 
				string are returned. The string should be in lower case.
	
	Exit:	function result = result code.
			strings = handle to group name strings.
			numGroups = number of group names.
			
	The time parameter is in GetDateTime format (seconds since 01/01/04). It
	is interpreted by the server, not by the Mac. To avoid missing new groups
	due to differences between the server's time and the Mac's time, you
	should subtract a healthy amount from the time of your last new groups 
	check (e.g., 36 hours), then discard any duplicate groups returned.  
			
	The strings block contains the C-format group names. Group names are 
	truncated to 127 characters.
----------------------------------------------------------------------------*/

NntpErr NntpGetGroupNames (NntpStreamHandle stream, unsigned long time,
	char *statusFilter, Handle *strings, long *numGroups)
{
	TStreamHandle s;
	NetStreamHandle netStream;
	DateTimeRec timeRec;
	Handle text = nil;
	char *p, *pEnd, *q, *r, status;
	long n, len;
	NntpErr nntpErr;
	
	s = (TStreamHandle)stream;
	nntpErr = ReOpenConnection(s);
	if (nntpErr != nntpNoErr) return nntpErr;
	netStream = (**s).netStream;
	
	if (time == 0) {
		strcpy(gCommand, "LIST");
	} else {
		SecondsToDate(time, &timeRec);
		timeRec.year = timeRec.year % 100;
		sprintf(gCommand, "NEWGROUPS %02d%02d%02d %02d%02d%02d", 
			timeRec.year, timeRec.month, timeRec.day,
			timeRec.hour, timeRec.minute, timeRec.second);
	}
	
	gErr = NetCommand(netStream, gCommand, &gResponseCode, gResponse);
	if (gErr != noErr) goto exit1;
	if (gResponseCode != (time == 0 ? 215 : 231)) goto exit2;
	
	gErr = NetGetText(netStream, &text);
	if (gErr != noErr) goto exit1;

	MyHLock(text);
	p = q = *text;
	pEnd = p + MyGetHandleSize(text);
	n = 0;
	while (p < pEnd) {
		gErr = NetGiveTime(false);
		if (gErr != noErr) goto exit3;
		r = p;
		while (*r != ' ' && *r != CR) r++;
		len = r-p;
		if (len > 127) len = 127;
		while (*r == ' ') r++;
		while (*r != ' ' && *r != CR) r++;
		while (*r == ' ') r++;
		while (*r != ' ' && *r != CR) r++;
		while (*r == ' ') r++;
		status = tolower(*r);
		while (*r != CR) r++;
		if (strchr(statusFilter, status) == nil) {
			*(p+len) = 0;
			strcpy(q, p);
			q += len+1;
			n++;
		}
		p = r+1;
	}
	len = q-*text;
	MyHUnlock(text);
	
	MySetHandleSize(text, len);
	*strings = text;
	*numGroups = n;
	SetLastTransactionTime(s);

	return nntpNoErr;
	
exit1:

	MyDisposeHandle(text);
	(**s).netStream = nil;
	return nntpOSErr;
	
exit2:
	
	MyDisposeHandle(text);
	return nntpServerErr;
	
exit3:

	MyDisposeHandle(text);
	return nntpOSErr;
}



/*----------------------------------------------------------------------------
	NntpGetGroupInfo 
	
	Get info for a single group.
	
	Entry:	stream = handle to stream.
			group = C-format group name.
	
	Exit:	function result = result code.
			first = first article number in group.
			last = last article number in group.
			count = server's estimate of number of articles in the group.
			
	This function returns nntpNoSuchGroupErr if the group does not exist.
	
	If the newConnection stream option is set, the current connection to 
	the news server is closed and a new one is opened before getting the 
	group information. You need to do this with some kinds of news servers
	(e.g., the reference implementation server). This close/reopen is
	not done, however, if the previous call to NntpGetGroupInfo was 
	within 10 seconds ago. This prevents rapid sequences of calls to
	NntpGetGroupInfo from closing and reopening.
----------------------------------------------------------------------------*/

NntpErr NntpGetGroupInfo (NntpStreamHandle stream, char *group, 
	long *first, long *last, long *count)
{
	TStreamHandle s;
	NetStreamHandle netStream;
	NntpErr nntpErr;
	char *p;
	static long prevTick = 0;
	
	s = (TStreamHandle)stream;
	
	if ((**s).options.newConnection && TickCount() > prevTick + 10*60) {
		netStream = (**s).netStream;
		if (netStream != nil) {
			NetClose(netStream);
			(**s).netStream = nil;
		}
	}
	
	nntpErr = ReOpenConnection(s);
	if (nntpErr != nntpNoErr) return nntpErr;
	netStream = (**s).netStream;
	
	*(**s).curGroup = 0;
	
	sprintf(gCommand, "GROUP %.127s", group);
	gErr = NetCommand(netStream, gCommand, &gResponseCode, gResponse);
	if (gErr != noErr) goto exit1;
	if (gResponseCode == 502) {
		*count = 0;
		*first = 1;
		*last = 0;
	} else if (gResponseCode == 211) {
		p = gResponse;
		CrackNum(&p);
		*count = CrackNum(&p);
		*first = CrackNum(&p);
		*last = CrackNum(&p);
	} else if (gResponseCode == 411) {
		return nntpNoSuchGroupErr;
	} else {
		return nntpServerErr;
	}

	SetCurrentGroup(s, group);
	SetLastTransactionTime(s);
	prevTick = TickCount();
	
	return nntpNoErr;
	
exit1:

	(**s).netStream = nil;
	return nntpOSErr;
}



/*----------------------------------------------------------------------------
	NntpGetMultipleGroupInfo 
	
	Get info for multiple groups.
	
	Entry:	stream = handle to stream.
			info = handle to array of group info.
			numGroups = number of groups in group info array.
			strings = handle to group name strings.
	
	Exit:	function result = result code.
			
	For each group in the array, the following information is returned in
	the group info array:
	
		first = first article number in group.
		last = last article number in group.
		count = estimate of number of articles in the group.
		ok = true if other info is valid, false if group does not exist
			or some other server error occured.
	
	If the newConnection stream option is set, the current connection to 
	the news server is closed and a new one is opened before getting the 
	group information. You need to do this with some kinds of news servers
	(e.g., the reference implementation server).
	
	If the batchedCmds stream option is set, multiple GROUP commands are 
	sent to the news server in a batch, and the responses are processed as
	they arrive. If the batchedCmds stream option is not set, the GROUP 
	commands are sent one at a time. Batched commands are much faster, but
	they do not work with all servers.
----------------------------------------------------------------------------*/

NntpErr NntpGetMultipleGroupInfo (NntpStreamHandle stream, NntpGroupInfoHandle info,
	long numGroups, Handle strings)
{
	TStreamHandle s;
	NetStreamHandle netStream;
	NntpErr nntpErr;
	NntpGroupInfoPtr x, xEnd;
	Boolean savedNewConnectionOption;
	long cmdBufSize;
	Handle cmdBuf;
	char *p;
	short len;
	short state1, state2;
	
	state1 = MyHGetState(info);
	state2 = MyHGetState(strings);
	
	if (numGroups == 0) return nntpNoErr;
	
	s = (TStreamHandle)stream;
	
	if ((**s).options.newConnection) {
		netStream = (**s).netStream;
		if (netStream != nil) {
			NetClose(netStream);
			(**s).netStream = nil;
		}
	}
	
	nntpErr = ReOpenConnection(s);
	if (nntpErr != nntpNoErr) return nntpErr;
	netStream = (**s).netStream;
	
	*(**s).curGroup = 0;
	
	if ((**s).options.batchedCmds) {
	
		cmdBufSize = 0;
		xEnd = *info + numGroups;
		for (x = *info; x < xEnd; x++) 
			cmdBufSize += 7 + strlen(*strings + x->offset);
		gErr = MyNewHandle(cmdBufSize, &cmdBuf);
		if (gErr != noErr) return nntpOSErr;
		p = *cmdBuf;
		xEnd = *info + numGroups;
		for (x = *info; x < xEnd; x++) {
			BlockMoveData("GROUP ", p, 6);
			p += 6;
			len = strlen(*strings + x->offset);
			BlockMoveData(*strings + x->offset, p, len);
			p += len;
			*p++ = CR;
		}
		gInfo = info;
		gInfoIndex = 0;
		gGroupInfoServerError = false;
		gErr = NetBatchedCommands(netStream, cmdBuf, DoOneGroupCmdResponse);
		MyDisposeHandle(cmdBuf);
		if (gErr != noErr) goto exit1;
		if (gGroupInfoServerError) return nntpServerErr;
	
	} else {
	
		savedNewConnectionOption = (**s).options.newConnection;
		(**s).options.newConnection = false;
		MyHLock(info);
		MyHLock(strings);
		xEnd = *info + numGroups;
		for (x = *info; x < xEnd; x++) {
			nntpErr = NntpGetGroupInfo(stream, *strings + x->offset,
				&x->first, &x->last, &x->count);
			if (nntpErr == nntpNoErr) {
				x->ok = true;
			} else if (nntpErr == nntpNoSuchGroupErr || nntpErr == nntpServerErr) {
				x->ok = false;
			} else {
				goto exit2;
			}
		}
		MyHSetState(info, state1);
		MyHSetState(strings, state2);
		(**s).options.newConnection = savedNewConnectionOption;
	
	}

	SetLastTransactionTime(s);
	
	return nntpNoErr;
	
exit1:

	MyHSetState(info, state1);
	MyHSetState(strings, state2);
	(**s).netStream = nil;
	return nntpOSErr;

exit2:

	MyHSetState(info, state1);
	MyHSetState(strings, state2);
	(**s).options.newConnection = savedNewConnectionOption;
	return nntpErr;

}



/*----------------------------------------------------------------------------
	NntpGetHeaders 
	
	Get selected article headers.
	
	Entry:	stream = handle to stream.
			group = group name.
			first = first article number.
			last = last article number.
			header = header name.
			pattern = search string.
			buildXPAT = pointer to function to build XPAT pattern
				for search string. Ignored if pattern nil or empty.
			matchPattern = pointer to function to compare search string
				to target string. Ignored if pattern nil or empty.
			
	Exit:	function result = result code.
			info = handle to array of header info.
			strings = handle to header strings.
			numHeaders = number of headers.
	
	If the pattern parameter is nil or the empty string, all headers in the 
	range [first,last] are returned. 
	
	If the pattern parameter is not nil or empty, only headers in the range
	[first,last] which match the pattern are returned.
			
	The strings block contains the C-format header strings, one after another.
	The offset fields in the header info array contain offsets into this
	block. Header strings are truncated to 255 characters, and have leading
	and trailing blanks stripped.
	
	All returned elements of the header info array have article numbers in 
	the range [first,last]. The header info array is sorted by article number, 
	with any duplicates eliminated.
	
	If the useXPAT stream option is set, and if a search pattern is specified,
	the buildXPAT callback function is called to build the regular expression
	XPAT pattern. An XPAT command is then sent to the server to get the 
	matching headers.
	
	If the useXPAT stream option is not set, or if the server does not support
	the XPAT command, all the headers are fetched from the server, and the
	matchPattern callback function is called to extract the matching headers.
----------------------------------------------------------------------------*/

NntpErr NntpGetHeaders (NntpStreamHandle stream, char *group, long first, long last,
	char *header, char *pattern,
	void (*buildXPAT)(char *pattern, char *regExp, short regExpLen),
	Boolean (*matchPattern)(char *pattern, char *string),
	NntpHeaderInfoHandle *info, Handle *strings, long *numHeaders)
{
	TStreamHandle s;
	NetStreamHandle netStream;
	NntpErr nntpErr;
	Handle text = nil;
	NntpHeaderInfoHandle theInfo = nil;
	NntpHeaderInfoPtr x;
	char *p, *pEnd, *q, *r, *t;
	long n, len, number, prevNumber;
	char xpatCmd[1000];
	Boolean mustFilter;
	Boolean mustSort = false;
	long numToMove;
	
	s = (TStreamHandle)stream;
	nntpErr = ReOpenConnection(s);
	if (nntpErr != nntpNoErr) return nntpErr;
	netStream = (**s).netStream;
	
	if (!MyStrEqual(group, (**s).curGroup)) {
		*(**s).curGroup = 0;
		sprintf(gCommand, "GROUP %.127s", group);
		gErr = NetCommand(netStream, gCommand, &gResponseCode, gResponse);
		if (gErr != noErr) goto exit1;
		if (gResponseCode == 411) return nntpNoSuchGroupErr;
		if (gResponseCode != 211) return nntpServerErr;
		SetCurrentGroup(s, group);
	}
		
	while (true) {
		if (pattern == nil || *pattern == 0 || !(**s).options.useXPAT || !(**s).serverHasXPAT) {
			sprintf(gCommand, "XHDR %.127s %ld-%ld", header, first, last);
			gErr = NetCommand(netStream, gCommand, &gResponseCode, gResponse);
			if (gErr != noErr) goto exit1;
			if (gResponseCode != 221) return nntpServerErr;
			mustFilter = pattern != nil && *pattern != 0;
			break;
		} else {
			sprintf(xpatCmd, "XPAT %.127s %ld-%ld ", header, first, last);
			len = strlen(xpatCmd);
			(*buildXPAT)(pattern, xpatCmd + len, 1000 - len);
			len = strlen(xpatCmd);
			if (len > 255) len = 255;
			BlockMoveData(xpatCmd, gCommand, len);
			*(gCommand + len) = 0;
			gErr = NetCommand(netStream, xpatCmd, &gResponseCode, gResponse);
			if (gErr != noErr) goto exit1;
			if (gResponseCode == 500) {
				(**s).serverHasXPAT = false;
				continue;
			}
			if (gResponseCode != 221) return nntpServerErr;
			mustFilter = false;
			break;
		}
	}
	
	gErr = NetGetText(netStream, &text);
	if (gErr != noErr) goto exit1;

	pEnd = *text + MyGetHandleSize(text);
	n = 0;
	for (p = *text; p < pEnd; p++) if (*p == CR) n++;
	
	gErr = MyNewHandle(n*sizeof(NntpHeaderInfo), &theInfo);
	if (gErr != noErr) goto exit3;

	MyHLock(text);
	MyHLock(theInfo);
	x = *theInfo;
	p = q = *text;
	pEnd = *text + MyGetHandleSize(text);
	n = 0;
	while (p < pEnd) {
		number = CrackNum(&p);
		if (number < first || number > last) {
			while (*p != CR) p++;
			p++;
			continue;
		}
		while (*p == ' ') p++;
		r = p;
		while (*r != CR) r++;
		t = r-1;
		while (t >= p && *t == ' ') t--;
		len = t-p+1;
		if (len > 255) len = 255;
		*(p+len) = 0;
		if (!mustFilter || (*matchPattern)(pattern, p)) {
			if (n > 0 && number <= prevNumber) mustSort = true;
			prevNumber = number;
			x->number = number;
			x->offset = q - *text;
			strcpy(q, p);
			q += len+1;
			n++;
			x++;
		}
		p = r+1;
	}
	if (mustSort) {
		gErr = FastQSort(*theInfo, n, sizeof(NntpHeaderInfo), 
			(short(*)(void*,void*))SortHeadersCompare);
		if (gErr != noErr) goto exit3;
		if (n > 1) {
			x =  *theInfo;
			for (numToMove = (n-1)*sizeof(NntpHeaderInfo); 
				numToMove > 0; 
				numToMove -= sizeof(NntpHeaderInfo))
			{
				if (x->number == (x+1)->number) {
					BlockMoveData(x+1, x, numToMove);
					n--;
				} else {
					x++;
				}
			}
		}
	}
	len = q - *text;
	MyHUnlock(text);
	MyHUnlock(theInfo);
	
	MySetHandleSize(text, len);
	MySetHandleSize(theInfo, n*sizeof(NntpHeaderInfo));
	*info = theInfo;
	*strings = text;
	*numHeaders = n;
	SetLastTransactionTime(s);
	
	return nntpNoErr;
	
exit1:

	MyDisposeHandle(text);
	MyDisposeHandle(theInfo);
	(**s).netStream = nil;
	return nntpOSErr;
	
exit2:
	
	MyDisposeHandle(text);
	MyDisposeHandle(theInfo);
	return nntpServerErr;
	
exit3:

	MyDisposeHandle(text);
	MyDisposeHandle(theInfo);
	return nntpOSErr;
}



/*----------------------------------------------------------------------------
	NntpGetArticle 
	
	Get an article.
	
	Entry:	stream = handle to stream.
			group = group name, or nil if fetching by message id.
			number = article number. Ignored if fetching by message id.
			id = message id string, including < and > delimiters. Ignored
				if fetching by article number.
			part = which part of the article to get:
				"ARTICLE": full article text, header and body.
				"HEAD": only article header.
				"BODY": only article body.
			truncateIfAttachedFile = true to truncate the article if it
				contains an attached file. The BinHex or uuencode text
				for the attached file is not returned.
			requireEncodedTextBeginLIne = true if BinHex or uuencode text
				must include special "begin" flag line.
			
	Exit:	function result = result code.
			text = handle to article text, including header lines, with CR 
				line terminators.
			*attachedFile = true if article contains an attached file which
				was truncated.
----------------------------------------------------------------------------*/

NntpErr NntpGetArticle (NntpStreamHandle stream, char *group, long number, 
	char *id, char *part, Boolean truncateIfAttachedFile, 
	Boolean requireEncodedTextBeginLine, Handle *text,
	Boolean *attachedFile)
{
	TStreamHandle s;
	NetStreamHandle netStream;
	NntpErr nntpErr;
	
	s = (TStreamHandle)stream;
	nntpErr = ReOpenConnection(s);
	if (nntpErr != nntpNoErr) return nntpErr;
	netStream = (**s).netStream;
	
	gRequireEncodedTextBeginLine = requireEncodedTextBeginLine;
	
	if (group != nil && !MyStrEqual(group, (**s).curGroup)) {
		*(**s).curGroup = 0;
		sprintf(gCommand, "GROUP %.127s", group);
		gErr = NetCommand(netStream, gCommand, &gResponseCode, gResponse);
		if (gErr != noErr) goto exit;
		if (gResponseCode == 411) return nntpNoSuchGroupErr;
		if (gResponseCode != 211) return nntpServerErr;
		SetCurrentGroup(s, group);
	}

	if (group != nil) {
		sprintf(gCommand, "%s %ld", part, number);
	} else {
		sprintf(gCommand, "%s %.127s", part, id);
	}
	gErr = NetCommand(netStream, gCommand, &gResponseCode, gResponse);
	if (gErr != noErr) goto exit;
	if (gResponseCode == 423 || gResponseCode == 430) return nntpNoSuchArticleErr;
	if (gResponseCode != 220 && gResponseCode != 221 &&
		gResponseCode != 222) return nntpServerErr;
	
	if (truncateIfAttachedFile) {
		gErr = NetGetTextWithTruncation(netStream, CheckForAttachedFile, text, attachedFile);
		if (gErr != noErr) goto exit;
		if (*attachedFile) (**s).netStream = nil;
	} else {
		gErr = NetGetText(netStream, text);
		if (gErr != noErr) goto exit;
	}

	SetLastTransactionTime(s);
	
	return nntpNoErr;
	
exit:

	(**s).netStream = nil;
	return nntpOSErr;
}



/*----------------------------------------------------------------------------
	NntpGetChunkyArticle
	
	Get an article in chunks.
	
	Entry:	stream = handle to stream.
			group = group name, or nil if fetching by message id.
			number = article number. Ignored if fetching by message id.
			id = message id string, including < and > delimiters. Ignored
				if fetching by article number.
			part = which part of the article to get:
				"ARTICLE": full article text, header and body.
				"HEAD": only article header.
				"BODY": only article body.
			chunkFunction = pointer to chunk processing function
	
	Exit:	function result = error code.
			*aborted = true if operation was aborted.
	
	The chunk processing function is called as the article text is received 
	from the server. It must be declared as follows:
	
	OSErr ChunkFunction (Ptr t, long tLen)
	
	Entry:	t = pointer to chunk.
			tLen = length of chunk.
			
	Exit:	function result = error code.
			*abort = true to abort the operation.
	
	The text passed to the chunk function is raw. It contains CRLF line 
	terminators and doubled leading ".." characters. The chunks do not
	necessarily end on CRLF line boundaries. The terminating "." character
	on a line by itself is *not* passed to the chunk function, however.
----------------------------------------------------------------------------*/

NntpErr NntpGetChunkyArticle (NntpStreamHandle stream, char *group, long number, 
	char *id, char *part, NntpChunkFunction chunkFunction,
	Boolean *aborted)
{
	TStreamHandle s;
	NetStreamHandle netStream;
	NntpErr nntpErr;
	
	s = (TStreamHandle)stream;
	nntpErr = ReOpenConnection(s);
	if (nntpErr != nntpNoErr) return nntpErr;
	netStream = (**s).netStream;
	
	if (group != nil && !MyStrEqual(group, (**s).curGroup)) {
		*(**s).curGroup = 0;
		sprintf(gCommand, "GROUP %.127s", group);
		gErr = NetCommand(netStream, gCommand, &gResponseCode, gResponse);
		if (gErr != noErr) goto exit;
		if (gResponseCode == 411) return nntpNoSuchGroupErr;
		if (gResponseCode != 211) return nntpServerErr;
		SetCurrentGroup(s, group);
	}

	if (group != nil) {
		sprintf(gCommand, "%s %ld", part, number);
	} else {
		sprintf(gCommand, "%s %.127s", part, id);
	}
	gErr = NetCommand(netStream, gCommand, &gResponseCode, gResponse);
	if (gErr != noErr) goto exit;
	if (gResponseCode == 423 || gResponseCode == 430) return nntpNoSuchArticleErr;
	if (gResponseCode != 220 && gResponseCode != 221 &&
		gResponseCode != 222) return nntpServerErr;
	
	gErr = NetGetChunkyText(netStream, chunkFunction, aborted);
	if (gErr != noErr) goto exit;
	
	if (*aborted) (**s).netStream = nil;

	SetLastTransactionTime(s);
	
	return nntpNoErr;
	
exit:

	(**s).netStream = nil;
	return nntpOSErr;
}



/*----------------------------------------------------------------------------
	NntpPostArticle 
	
	Post an article.
	
	Entry:	stream = handle to stream.
			text = handle to article text, including header lines, 
				with CR line terminators.
				Warning: the memory block is modified by the function.
				The memory block must be unlocked and nonpurgeable.
			
	Exit:	function result = result code.
			postIndeterminate = true if entire article sent, but error occured
				or user canceled before final server response received. The
				article may or may not have been posted successfully.
----------------------------------------------------------------------------*/

NntpErr NntpPostArticle (NntpStreamHandle stream, Handle text, Boolean *postIndeterminate)
{
	TStreamHandle s;
	NetStreamHandle netStream;
	NntpErr nntpErr;
	
	*postIndeterminate = false;
	
	s = (TStreamHandle)stream;
	nntpErr = ReOpenConnection(s);
	if (nntpErr != nntpNoErr) return nntpErr;
	netStream = (**s).netStream;
	
	strcpy(gCommand, "POST");
	gErr = NetCommand(netStream, gCommand, &gResponseCode, gResponse);
	if (gErr != noErr) goto exit1;
	if (gResponseCode != 340) return nntpServerErr;
	
	gErr = NetSendText(netStream, text);
	if (gErr != noErr) goto exit1;
	
	*postIndeterminate = true;
	
	gErr = NetGetExtraResponse(netStream, &gResponseCode, gResponse);
	if (gErr != noErr) goto exit1;
	*postIndeterminate = false;
	if (gResponseCode != 240) return nntpServerErr;

	SetLastTransactionTime(s);
	
	return nntpNoErr;
	
exit1:

	(**s).netStream = nil;
	return nntpOSErr;
}



/*----------------------------------------------------------------------------
	NntpAuthorize 
	
	Authorize user.
	
	Entry:	stream = handle to stream.
			username = username.
			password = password.
			
	Exit:	function result = result code.
----------------------------------------------------------------------------*/

NntpErr NntpAuthorize (NntpStreamHandle stream, char *username, char *password)
{
	TStreamHandle s;
	NetStreamHandle netStream;
	NntpErr nntpErr;
	
	s = (TStreamHandle)stream;
	nntpErr = ReOpenConnection(s);
	if (nntpErr != nntpNoErr) return nntpErr;
	netStream = (**s).netStream;

	nntpErr = Authorize(netStream, username, password);
	if (nntpErr != nntpNoErr) goto exit;
	
	SetLastTransactionTime(s);
	
	return nntpNoErr;
	
exit:

	(**s).netStream = nil;
	return nntpErr;
}



/*----------------------------------------------------------------------------
	NntpGetHello 
	
	Get the server hello message.
	
	Entry:	stream = handle to stream.
			
	Exit:	function result = result code (always nntpNoErr).
			msg = hello message.
----------------------------------------------------------------------------*/

NntpErr NntpGetHello (NntpStreamHandle stream, CStr255 msg)
{
	TStreamHandle s;
	
	s = (TStreamHandle)stream;
	strcpy(msg, (**s).helloMsg);
	return nntpNoErr;
}



/*----------------------------------------------------------------------------
	NntpGetHelp 
	
	Get the server help text.
	
	Entry:	stream = handle to stream.
			
	Exit:	function result = result code.
			text = handle to help text, with CR-terminated lines.
----------------------------------------------------------------------------*/

NntpErr NntpGetHelp (NntpStreamHandle stream, Handle *text)
{
	TStreamHandle s;
	NetStreamHandle netStream;
	NntpErr nntpErr;
	
	s = (TStreamHandle)stream;
	nntpErr = ReOpenConnection(s);
	if (nntpErr != nntpNoErr) return nntpErr;
	netStream = (**s).netStream;
	
	strcpy(gCommand, "HELP");
	gErr = NetCommand(netStream, gCommand, &gResponseCode, gResponse);
	if (gErr != noErr) goto exit1;
	if (gResponseCode != 100) return nntpServerErr;
	
	gErr = NetGetText(netStream, text);
	if (gErr != noErr) goto exit1;

	SetLastTransactionTime(s);
	
	return nntpNoErr;
	
exit1:

	(**s).netStream = nil;
	return nntpOSErr;
}



/*----------------------------------------------------------------------------
	NntpGetIPAddr 
	
	Get server IP address.
		
	Entry:	stream = handle to stream.
			
	Exit:	function result = result code (always nntpNoErr).
			ipAddr = IP address of news server.
----------------------------------------------------------------------------*/

NntpErr NntpGetIPAddr (NntpStreamHandle stream, unsigned long *ipAddr)
{
	TStreamHandle s;
	
	s = (TStreamHandle)stream;
	*ipAddr = (**s).addr;
	return nntpNoErr;
}



/*----------------------------------------------------------------------------
	NntpGetServerErrInfo 
	
	Get server error information.
	
	Exit:	cmd = C-format server command.
			num = server error number.
			msg = C-format server error message.
----------------------------------------------------------------------------*/

void NntpGetServerErrInfo (CStr255 cmd, long *num, CStr255 msg)
{
	strcpy(cmd, gCommand);
	*num = gResponseCode;
	strcpy(msg, gResponse);
}



/*----------------------------------------------------------------------------
	NntpGetOSErr 
	
	Get the OS error code.
	
	Exit:	function result = OS error code.
----------------------------------------------------------------------------*/

OSErr NntpGetOSErr (void)
{
	return gErr;
}