/* This is a comparison of several sorting functions. After running
   this program, look in the output file (see the define for OUTPUT),
   or at the standard output to see how long each sort took.

   While all the sorting functions are standard ANSI C, and don't
   include any ifdef's, this file has to interact with host
   environments which aren't neccessarily ANSI compatible. I've
   therefore placed all of the system dependent functions called
   from this file into the file "sortsys.c". */

#include <assert.h>
#include <limits.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include "sortlib.h"
#include "tsortlib.h"
#include "sortsys.h"

/*--------------------------------------------------------------------------*/
/* various definitions */
/*--------------------------------------------------------------------------*/

/* absolute maximum number of data items */
#define DATA_NELEM_MAX		LONG_MAX

/* number of items to sort */
#ifndef DATA_NELEM
  #define DATA_NELEM		(8196)
#endif

/* number of items to sort for small files */
#define DATA_NELEM_SMALL	(16)

/* number of seconds to run each sort */
#define SORT_TIME		(5 * CLOCKS_PER_SEC)

/* name of output file */
#ifndef SORT_OUTPUT
  #define SORT_OUTPUT "sortout" 
#endif

/* default strings describing the system the program is running on */

#ifndef MC_NAME
  #define MC_NAME "Unknown"
#endif

#ifndef OS_NAME
  #define OS_NAME "Unknown"
#endif

#ifndef CC_NAME
  #define CC_NAME "Unknown"
#endif

/*--------------------------------------------------------------------------*/
/* types */
/*--------------------------------------------------------------------------*/

/* indexes to data arrays */
enum {
  DATA_RANDOM,
  DATA_SORTED,
  DATA_REVERSED,
  DATA_LAST
};

/* type of data to sort; if you change this, you'll also need to change
   the file "tsort_private.h" */
typedef int data_type;

/* pointer to a comparison function */
typedef int (*CFPTR)(const void *a, const void *b);

/* pointer to a sorting function */
typedef void (*SFPTR)(void *base, size_t n, size_t size, CFPTR compar);

/*--------------------------------------------------------------------------*/
/* globals */
/*--------------------------------------------------------------------------*/

/* version string */
static const char *const version =
  "Sort Library, Version 2.0, Copyright (c) 1990-1994, Ari Halberstadt";

/* pointer to first arg */
static const char *progname;

/* number of items to sort */
static long data_nelem_max;

/* stream to output results to */
static FILE *sortout;

/*--------------------------------------------------------------------------*/
/* test utilities */
/*--------------------------------------------------------------------------*/

/* fill data with random integers */
static void randomize(data_type *data, long n)
{
  while (n-- > 0)
    data[n] = rand() % INT_MAX;
}

/* compare two integers */
static data_type compare(const void *a, const void *b)
{
  return(*(const data_type *) a - *(const data_type *) b);
}

/*--------------------------------------------------------------------------*/
/* printing test results */
/*--------------------------------------------------------------------------*/

/* print one character */
static void prchar(char c)
{
  putc(c, stdout);
  if (sortout)
    putc(c, sortout);
}

/* print formatted output */
static void prfmt(const char *fmt, ...)
{
  va_list ap;

  va_start(ap, fmt);
  vfprintf(stdout, fmt, ap);
  if (sortout)
    vfprintf(sortout, fmt, ap);
  va_end(ap);
}

/* print table header */
static void printhdr(void)
{
  int i;
  
  prfmt("\n  %-8s %-8s %12s %12s %12s %12s\n",
  	"sort", "order", "nelem", "seconds", "sec/elem", "qsort/sort");
  prfmt("  ");
  for (i = 0; i < 69; i++)
    prchar('-');
  prchar('\n');
}

/* print run info */
static void printrun(const char *prompt_name, const char *prompt_order,
  long total_nelem, clock_t total_elapsed,
  long qsort_nelem, clock_t qsort_elapsed)
{
  double total_seconds;
  double qsort_seconds;
  
  total_seconds = (double) total_elapsed / CLOCKS_PER_SEC;
  qsort_seconds = (double) qsort_elapsed / CLOCKS_PER_SEC;
  prfmt("  %-8s %-8s %12ld %12.3f %12.3e %12.3e\n",
    prompt_name,
    prompt_order,
    total_nelem,
    total_seconds,
    total_seconds / total_nelem,
    (qsort_seconds / qsort_nelem) / (total_seconds / total_nelem));
}

/*--------------------------------------------------------------------------*/
/* functions for running and verifying the tests */
/*--------------------------------------------------------------------------*/

/* verify that the data are in ascending order and match the sorted data */
static bool verify(const data_type *data, const data_type *sorted,
  long data_nelem)
{
  const long context = 16;
  long i, before, after;
  bool valid;
  
  for (i = 1; i < data_nelem && data[i-1] <= data[i]; i++)
    ;
  valid = (i >= data_nelem);
  if (valid && data_nelem == data_nelem_max) {
    for (i = 0; i < data_nelem && data[i] == sorted[i]; i++)
      ;
    valid = (i >= data_nelem);
  }
  if (! valid) {
    before = (i - context <= 0 ? 0 : i - context);
    after = (i + context >= data_nelem ? data_nelem : i + context);
    prfmt("\tSort failed at index %ld. Context (%ld before, %ld after):\n",
    	i, i - before, after - i);
    prfmt("\t  data:   ");
    for (i = before; i < after; i++)
    	prfmt("%d ", data[i]);
    prfmt("\n");
    if (data_nelem == data_nelem_max) {
      prfmt("\t  sorted: ");
      for (i = before; i < after; i++)
	prfmt("%d ", sorted[i]);
      prfmt("\n");
    }
  }
  return(valid);
}

/* time and verify the given sort function */
static void sort(
  SFPTR function_sort,
  CFPTR function_compare,
  const char *prompt_name,
  const char *prompt_order,
  const data_type *data_unsorted,
  const data_type *data_sorted,
  data_type *data_copy,
  long data_nelem,
  long *total_nelem,
  clock_t *total_elapsed)
{
  clock_t elapsed;
  clock_t start;
  clock_t stop;
  long nelem;
  long i;
  
  nelem = 0;
  elapsed = 0;
  do {
  
    /* copy unsorted data */
    for (i = 0; i < data_nelem; i++)
      data_copy[i] = data_unsorted[i];
  
    /* sort */
    profile_start();
    start = clock();
    function_sort(data_copy, data_nelem, sizeof(data_type), function_compare);
    stop = clock();
    profile_stop();
    
    /* tally */
    nelem += data_nelem;
    elapsed += stop - start;
    
    /* verify sort */
    if (! verify(data_copy, data_sorted, data_nelem)) {
      prfmt("\tSort failed: %s, %ld %s integers\n",
        prompt_name, data_nelem, prompt_order);
      break;
    }
  
    /* only run once for small data sets */
    if (data_nelem <= DATA_NELEM_SMALL)
      break;
      
  } while (elapsed < SORT_TIME);
  
  /* return results */
  if (total_nelem)
    *total_nelem = nelem;
  if (total_elapsed)
    *total_elapsed = elapsed;
}

/* return name of function */
static const char *fname(SFPTR fptr)
{
  static struct {
    SFPTR fptr;		/* pointer to function */
    const char *name;	/* name of function */
  } names[] = {
    { qsort,	"qsort",	},
    { qksort,	"qksort",	},
    { hpsort,	"hpsort",	},
    { mgsort,	"mgsort",	},
    { shsort,	"shsort",	},
    { insort,	"insort",	},
    { tqksort,	"tqksort",	},
    { thpsort,	"thpsort",	},
    { tmgsort,	"tmgsort",	},
    { tshsort,	"tshsort",	},
    { tinsort,	"tinsort",	},
    { NULL,	"",		},
  };
  int i;
  
  for (i = 0; names[i].fptr && names[i].fptr != fptr; i++)
    ;
  return(names[i].name);
}

struct test {		/* function to test */
  SFPTR fptr;		/* pointer to function */
  long nelem;		/* number of items */
  long total_nelem;	/* total number of items sorted */
  clock_t total_elapsed;/* total elapsed time */
};

/* comparison function used for sorting the results */
static int compare_test(const void *a, const void *b)
{
  const struct test *testa = (const struct test *) a;
  const struct test *testb = (const struct test *) b;
  double timea = (double) testa->total_elapsed / testa->total_nelem;
  double timeb = (double) testb->total_elapsed / testb->total_nelem;
  if (timea < timeb) return(-1);
  if (timea > timeb) return(1);
  return(0);
}

/* run the tests */
static void run(data_type *data_copy, data_type *data_array[])
{
  #define MIN(a)  ((a) < DATA_NELEM ? (a) : DATA_NELEM)
  /* This structure contains a list of the sort functions to test. Each
     entry also has the maximum values for each of the tests
     executed. Since different machines have different speeds
     and some implementations of qsort are better than others,
     it makes sense to have different maximum values. For instance,
     sorting 4096 presorted integers using THINK C's qsort is very
     slow, while on the NeXT it is very fast. You may have to tinker
     with this table to get reasonable execution times. */
  static const char *data_names[DATA_LAST] =
    { "random", "sorted", "reversed", };
  struct test tests_all[DATA_LAST][16] = {
    { /* random */
      { qsort,		DATA_NELEM,	0, 0, },
      { qksort,		DATA_NELEM,	0, 0, },
      { hpsort,		DATA_NELEM,	0, 0, },
      { mgsort,		DATA_NELEM,	0, 0, },
      { shsort,		MIN(16384),	0, 0, },
      { insort,		MIN(1024),	0, 0, },
      { tqksort,	DATA_NELEM,	0, 0, },
      { thpsort,	DATA_NELEM,	0, 0, },
      { tmgsort,	DATA_NELEM,	0, 0, },
      { tshsort,	MIN(16384),	0, 0, },
      { tinsort,	MIN(1024),	0, 0, },
      { NULL,		0,		0, 0, },
    },
    { /* sorted */
      #if THINK_C || __MWERKS__
        /* some quick sorts exhibit n^2 performance if array is sorted */
        { qsort,	MIN(512),	0, 0, },
      #else
        { qsort,	DATA_NELEM,	0, 0, },
      #endif
      { qksort,		DATA_NELEM,	0, 0, },
      { hpsort,		DATA_NELEM,	0, 0, },
      { mgsort,		DATA_NELEM,	0, 0, },
      { shsort,		MIN(16384),	0, 0, },
      { insort,		DATA_NELEM,	0, 0, },
      { tqksort,	DATA_NELEM,	0, 0, },
      { thpsort,	DATA_NELEM,	0, 0, },
      { tmgsort,	DATA_NELEM,	0, 0, },
      { tshsort,	MIN(16384),	0, 0, },
      { tinsort,	DATA_NELEM,	0, 0, },
      { NULL,		0,		0, 0, },
    },
    { /* reversed */
      #if THINK_C || __MWERKS__
        /* some quick sorts exhibit n^2 performance if array is reversed */
        { qsort,	MIN(512),	0, 0, },
      #else
        { qsort,	DATA_NELEM,	0, 0, },
      #endif
      { qksort,		DATA_NELEM,	0, 0, },
      { hpsort,		DATA_NELEM,	0, 0, },
      { mgsort,		DATA_NELEM,	0, 0, },
      { shsort,		MIN(16384),	0, 0, },
      { insort,		MIN(512),	0, 0, },
      { tqksort,	DATA_NELEM,	0, 0, },
      { thpsort,	DATA_NELEM,	0, 0, },
      { tmgsort,	DATA_NELEM,	0, 0, },
      { tshsort,	MIN(16384),	0, 0, },
      { tinsort,	MIN(512),	0, 0, },
      { NULL,		0,		0, 0, },
    },
  };
  
  long qsort_nelem;
  clock_t qsort_elapsed;
  struct test *test;
  int data_index;
  int test_index;
  int small_nelem;
  int ntests;
  
  /* keep compiler happy about initialization */
  qsort_nelem = 1;
  qsort_elapsed = 0;

  /* estimate time to run all tests */
  ntests = 0;
  for (data_index = 0; data_index < DATA_LAST; data_index++) {
    test = tests_all[data_index];
    for (test_index = 0; test[test_index].fptr; test_index++)
      ntests++;
  }
  prfmt("\nIt will take at least %d seconds to run all of the tests.\n",
  	(int) (ntests * SORT_TIME / CLOCKS_PER_SEC));
  
  /* run all groups of tests */
  for (data_index = 0; data_index < DATA_LAST; data_index++) {
  
    test = tests_all[data_index];
    
    /* print table header */
    prfmt("\nTesting %s data\n", data_names[data_index]);
    printhdr();
 
    /* run the tests on the data */
    for (test_index = 0; test[test_index].fptr; test_index++) {
    
      /* sort data */
      sort(test[test_index].fptr,
	   compare,
	   fname(test[test_index].fptr),
	   data_names[data_index],
	   data_array[data_index],
	   data_array[DATA_SORTED],
	   data_copy,
	   test[test_index].nelem,
	   &test[test_index].total_nelem,
	   &test[test_index].total_elapsed);

      /* remember values for system sort */
      if (test[test_index].fptr == qsort) {
        qsort_nelem = test[test_index].total_nelem;
        qsort_elapsed = test[test_index].total_elapsed;
      }
      
      /* print report */
      printrun(fname(test[test_index].fptr),
               data_names[data_index],
               test[test_index].total_nelem,
               test[test_index].total_elapsed,
               qsort_nelem,
               qsort_elapsed);
    }
   
    /* output sorted results */
    qksort(test, test_index, sizeof(struct test), compare_test);
  
    prfmt("\nSorted results:\n");
    printhdr();
    for (test_index = 0; test[test_index].fptr; test_index++) {
      printrun(fname(test[test_index].fptr),
	       data_names[data_index],
	       test[test_index].total_nelem,
	       test[test_index].total_elapsed,
	       qsort_nelem,
	       qsort_elapsed);
    }
  }
  
  /* test small lists; especially useful for catching boundary errors */
  prfmt("\nTesting small numbers of items (0 through %d)\n", DATA_NELEM_SMALL);
  test = tests_all[DATA_RANDOM];
  for (test_index = 0; test[test_index].fptr; test_index++) {
    
    /* test each function with various sizes of small sets of data */
    for (small_nelem = 0; small_nelem <= DATA_NELEM_SMALL; small_nelem++) {
      sort(test[test_index].fptr,
	   compare,
	   fname(test[test_index].fptr),
	   data_names[DATA_RANDOM],
	   data_array[DATA_RANDOM],
	   data_array[DATA_SORTED],
	   data_copy,
	   small_nelem,
	   NULL,
	   NULL);
    }
    
    /* skip to next function */
    while (test[test_index].fptr == test[test_index + 1].fptr)
      ++test_index;
  }
  
  prfmt("\n");
}

/* display information about the computer we're running on */
static void sysinfo(void)
{
  char mc_name[256];
  char os_name[256];
  char cc_name[256];
  
  strcpy(mc_name, MC_NAME);
  strcpy(os_name, OS_NAME);
  strcpy(cc_name, CC_NAME);
  
  #ifdef SYS_MACINTOSH
    sysinfomac(mc_name, os_name, cc_name);
  #endif

  prfmt("%-18s %s\n", "Hardware:", mc_name);
  prfmt("%-18s %s\n", "System:", os_name);
  prfmt("%-18s %s\n", "Compiler:", cc_name);
  prfmt("%-18s %ld\n", "sizeof(data_type)=", sizeof(data_type));
}

/* initialize the program */
static void initialize() 
{
  /* system dependent initializations */
  sysinit();
  
  /* open output file */
  sortout = fopen(SORT_OUTPUT, "w");

  /* seed random number generator */
  srand(clock());

  /* initialize profiler */
  profile_init();
  profile_stop();  
}

/* print usage and exit */
static void usage(void)
{
  fprintf(stderr, "Usage: %s [-s seed]\n", progname);
  exit(EXIT_FAILURE);
}

int main(int argc, char **argv)
{
  const long data_nelem = DATA_NELEM;/* number of elements in array */
  data_type *data_array[DATA_LAST];/* arrays of data to sort */
  data_type *data_copy;		   /* copy of data for sort result */
  int data_index;		   /* index to data_array */
  long i;			   /* loop counter */
  unsigned long bytes_needed;	   /* for estimating amount of memory needed */
  
  /* process command line arguments */
  progname = *argv;  
  while (--argc > 0 && (*++argv)[0] == '-') {
    char *s;    
    for (s = argv[0] + 1; *s != '\0'; s++) {
      switch (*s) {
      case 's':
	if (argc-- == 0)
	  usage();
	srand(atoi(*++argv));
	break;     
      default:
	fprintf(stderr, "%s: illegal option '%c'\n", progname, *s);
	usage();
      }
    }
  }
  
  /* do various initializations */
  initialize();
  
  /* print descriptive information */
  prfmt("%s\n", version);
  prfmt("Compiled on %s at %s\n", __DATE__, __TIME__);
  prfmt("\n");
  sysinfo();
  prfmt("\n");
    
  /* Create data. Once the data are created, we reuse them for each
     test, thus getting a fair comparison of the algorithms. */
  prfmt("Preparing data\n\n");
  
  prfmt("\tAllocating arrays\n");
  bytes_needed = 0;
  data_copy = malloc(sizeof(data_type) * data_nelem);
  for (data_index = 0; data_index < DATA_LAST; data_index++)
    data_array[data_index] = malloc(sizeof(data_type) * data_nelem);
  for (data_index = 0; data_index < DATA_LAST; data_index++) {
    if (! data_array[data_index])
      bytes_needed += sizeof(data_type) * data_nelem;
  }
  if (! data_copy)
    bytes_needed += sizeof(data_type) * data_nelem;
  if (bytes_needed > 0) {
    fprintf(stderr, "%s: can't allocate memory: approximately %lu additional bytes are needed.\n",
      progname, bytes_needed);
    exit(EXIT_FAILURE);
  }
  
  prfmt("\tCreating %ld random integers\n", data_nelem);
  randomize(data_array[DATA_RANDOM], data_nelem);
    
  prfmt("\tPresorting data\n");
  memcpy(data_array[DATA_SORTED], data_array[DATA_RANDOM],
  	sizeof(data_type) * data_nelem);
  qksort(data_array[DATA_SORTED], data_nelem, sizeof(data_type), compare);

  prfmt("\tReversing data\n");
  for (i = 0; i < data_nelem; i++)
    data_array[DATA_REVERSED][data_nelem - i - 1] = data_array[DATA_SORTED][i];
  
  #if SORT_NOMEM
    prfmt("\tSort algorithms will not use any memory, forcing use of\n");
    prfmt("\talternative algorithms for certain sorts\n");
  #endif
  
  prfmt("\nRunning tests. If you see any failure messages then it's time\n");
  prfmt("to fix something.\n");
  run(data_copy, data_array);
  prfmt("Finished all tests.\n");

  profile_dump();

  return(EXIT_SUCCESS);
}
